# -*- coding: utf-8 -*-
__RCSID__ = "$Id: $"

import datetime

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler

from IHEPDIRAC.TransferSystem.DB.TransferDB import TransRequestEntry

tmpGlobalStore = {}

global gTransferDB 

def initializeTransferRequestHandler(serviceInfo):
  """ initialize handler """

  gLogger.info("Initialize TransferRequestHandler.")

  from IHEPDIRAC.TransferSystem.DB.TransferDB import TransferDB

  global gTransferDB

  gTransferDB = TransferDB()

  return S_OK()

class TransferRequestHandler(RequestHandler):
  """
  This is for:
    * create a request to transfer data
  """

  def initialize(self):
    credDict = self.getRemoteCredentials()
    gLogger.info(credDict)
    self.user = credDict["username"]

    tmpGlobalStore.setdefault( self.user, 
                              {"endpoint":{}
                              } )

  types_serverIsOK = []
  def export_serverIsOK(self):
    return S_OK()

  types_getEndPoint = []
  def export_getEndPoint(self):
    ep = tmpGlobalStore[self.user]["endpoint"]

    return S_OK( ep )

  types_setEndPoint = [str, str]
  def export_setEndPoint(self, name, url):
    gLogger.info(name)
    gLogger.info(url)

    tmpGlobalStore[self.user]["endpoint"][name] = url

    return S_OK()

  def check_create_permission(self):
    credDict = self.getRemoteCredentials()
    if credDict["group"] in ["data_transfer"]:
      return True
    return False
    
  types_create = [ str, str, str, str ]
  def export_create(self, dataset, ep_from, ep_to, protocol):
    # check whether the user can create a request

    if not self.check_create_permission():
      return S_ERROR("The user can't create transfer request")
    entry = TransRequestEntry(username = self.user, 
                              dataset = dataset,
                              srcSE = ep_from,
                              dstSE = ep_to,
                              protocol = protocol,
                              status = "new",
                              submit_time = datetime.datetime.utcnow())
    gLogger.info("create an Entry:", entry)
    res = gTransferDB.insert_TransferRequest(entry)
    return res

  types_status = [ dict ]
  def export_status(self, condDict=None):
    """ This is give the status of the request db
    """
    res = gTransferDB.get_TransferRequest(condDict)
    return res

  types_statustotal = [ dict ]
  def export_statustotal(self, condDict=None):
    """ This is give the status of the request db
    """
    res = gTransferDB.get_TransferRequestTotal(condDict)
    return res

  types_statuslimit = [ dict, [list, str], [int, long], [int, long] ]
  def export_statuslimit(self, condDict=None, orderby=None, 
                               offset=None, limit=None):
    """ This is give the status of the request db
    """
    res = gTransferDB.get_TransferRequestWithLimit(condDict, orderby, offset, limit)
    return res

  types_show = [ dict ]
  def export_show(self, condDict):
    """ This is for query the file list.
    """
    res = gTransferDB.get_TransferFileList(condDict)
    return res

  types_delete = [ dict ]
  def export_delete(self, condDict):
    """ This will make the status to kill
         (new, transfer) --> kill
    """ 
    res = gTransferDB.delete_TransferFileList(condDict)
    return res

  types_delete_files_in_req = [ dict ]
  def export_delete_files_in_req(self, condDict):
    res = gTransferDB.delete_TransferFileListByReq(condDict)
    return res

  types_retransfer = [ dict ]
  def export_retransfer(self, condDict):
    """ This will make the status to new
         (kill, finish) --> new
    """ 
    res = gTransferDB.retransfer_TransferFileList(condDict)
    return res



import subprocess

from DIRAC import gLogger

class Rsync(object):
  def __init__(self):
    super(Rsync, self).__init__()

    # rsync could validate data by itself
#    self._localValidation = False

  def _downloadSingleFile(self, remotePath, localPath):
    gLogger.debug('rsync from %s to %s' % (remotePath, localPath))

    args = ['rsync', '-z', remotePath, localPath]
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    ret = p.returncode

    return ret == 0



from WebAppDIRAC.Lib.WebHandler import WebHandler, WErr
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
import tempfile
import datetime
import simplejson
from hashlib import md5
from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities import Time, List
from DIRAC.AccountingSystem.Client.ReportsClient   import ReportsClient

class TransferAccountingHandler(WebHandler):

    AUTH_PROPS = "authenticated"

    def __init__(self, *args, **kwargs ):
        super( TransferAccountingHandler, self ).__init__( *args, **kwargs )
        sessionData = self.getSessionData()
        self.user = sessionData['user'].get( 'username', '' )
        self.group = sessionData['user'].get( 'group', '' )
        self.vo = getVOForGroup( self.group )
        self.fc = FileCatalog( vo = self.vo )

        self.log.always(self.user)
        self.log.always(self.group)
        self.log.always(self.vo)
        self.log.always(self.fc)

    def web_getPilotList(self):
        '''
        self.log.debug(self.request.arguments)
        typeName = None
        if self.request.arguments.get("typeName", None):
            typeName = self.request.arguments["typeName"][0]

        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listReports(typeName)
        self.log.always(res)
        if res["OK"]:
            data = simplejson.dumps(res['Value'])
        self.log.always(data)
        data = eval(data)
        self.write({"result":data})

        '''
        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listReports(typeName)
        self.log.always(res)
        if res["OK"]:
           self.write({"result":res['Value']})

    def web_getSelectData(self):
        '''
        '''
        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listUniqueKeyValues(typeName)
        self.log.always(res)
        if res["OK"]:
            resd=res['Value'].keys()
        self.log.always('1111111111111111111111111111111111')
        S=res['Value']['Source']
        D=res['Value']['Destination']
        P=res['Value']['Protocol']
        F=res['Value']['FinalStatus']
        U=res['Value']['User']
        self.log.always(S)
        self.log.always('22222222222222222222222222222222')
        data=[]
        for i in resd:
            data.append(dict(value = i, text = i))
        self.finish({'success' : 'true', 'result' : data})

    def web_getSource(self):
        '''
        res['Value']={"Source": [], "Destination": [], "Protocol": [], "FinalStatus": [], "User": []}
        data[Source] data[Destination] data[Protocol] data[FinalStatus] data[User]
        list: s d p f u
        '''
        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listUniqueKeyValues(typeName)
        self.log.always(res)
        if res["OK"]:
            resd=res['Value'].keys()
        S=res['Value']['Source']
        self.log.always(S)
        data=[]
        for x in S:
            data.append(dict(value = x, text = x))
        self.finish({'success' : 'true', 'result' : data})


    def web_getDestination(self):
        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listUniqueKeyValues(typeName)
        self.log.always(res)
        if res["OK"]:
            resd=res['Value'].keys()
        S=res['Value']['Destination']
        self.log.always(S)
        data=[]
        for x in S:
            data.append(dict(value = x, text = x))
        self.finish({'success' : 'true', 'result' : data})



    def web_getProtocol(self):
        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listUniqueKeyValues(typeName)
        self.log.always(res)
        if res["OK"]:
            resd=res['Value'].keys()
        S=res['Value']['Protocol']
        self.log.always(S)
        data=[]
        for x in S:
            data.append(dict(value = x, text = x))
        self.finish({'success' : 'true', 'result' : data})



    def web_getFinalStatus(self):
        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listUniqueKeyValues(typeName)
        self.log.always(res)
        if res["OK"]:
            resd=res['Value'].keys()
        S=res['Value']['FinalStatus']
        self.log.always(S)
        data=[]
        for x in S:
            data.append(dict(value = x, text = x))
        self.finish({'success' : 'true', 'result' : data})



    def web_getUser(self):
        typeName="DataTransfer"
        rep=RPCClient("Accounting/ReportGenerator")
        res=rep.listUniqueKeyValues(typeName)
        self.log.always(res)
        if res["OK"]:
            resd=res['Value'].keys()
        S=res['Value']['User']
        self.log.always(S)
        data=[]
        for x in S:
            data.append(dict(value = x, text = x))
        self.finish({'success' : 'true', 'result' : data})


    def web_getPlotData(self):
        callback = {}
        retVal = self.__parseFormParams()
        if not retVal[ 'OK' ]:
            callback = {"success":"false", "error":retVal[ 'Message' ]}
            self.finish( callback )

        params = retVal[ 'Value' ]
        '''self.finish({'success' : 'true', 'result' : params})'''
        repClient = ReportsClient( rpcClient = RPCClient( "Accounting/ReportGenerator" ) )
        retVal = repClient.getReport(*params)
        if not retVal[ 'OK' ]:
            callback = {"success":"false", "error":retVal[ 'Message' ]}
            self.finish( callback )
        rawData = retVal[ 'Value' ]
        groupKeys = rawData[ 'data' ].keys()
        self.finish({'success' : 'true', 'result' : groupKeys})

    def web_generatePlot(self):
        callback = {}
        retVal = self.__parseFormParams()
        if not retVal[ 'OK' ]:
            callback = {"success":"false", "error":retVal[ 'Message' ]}
            self.finish( callback )

        params = retVal[ 'Value' ]
        repClient = ReportsClient( rpcClient = RPCClient( "Accounting/ReportGenerator" ) )
        retVal = repClient.generateDelayedPlot( *params )
        if not retVal[ 'OK' ]:
            callback = {"success":"false", "error":retVal[ 'Message' ]}
            self.finish( callback )
        rawData = retVal[ 'Value' ]
        '''groupKeys = rawData[ 'data' ].keys()'''
        self.log.always("11111111111111111111111111111111111111111111111111111111111")
        self.log.always(retVal[ 'Value' ][ 'plot' ])
        self.log.always("22222222222222222222222222222222222222222222222222222222222")
        self.finish({'success' : 'true', 'result' : rawData['plot']})




    def web_getPlotImg(self):
        """
        Get plot image
        """
        callback = {}
        if 'file' not in self.request.arguments:
            callback = {"success":"false", "error":"Maybe you forgot the file?"}
            self.finish( callback )
            return
        plotImageFile = str( self.request.arguments[ 'file' ][0] )

        if plotImageFile.find( ".png" ) < -1:
            callback = {"success":"false", "error":"Not a valid image!"}
            self.finish( callback )
            return

        transferClient = TransferClient( "Accounting/ReportGenerator" )
        tempFile = tempfile.TemporaryFile()
        retVal = transferClient.receiveFile( tempFile, plotImageFile )
        if not retVal[ 'OK' ]:
            callback = {"success":"false", "error":retVal[ 'Message' ]}
            self.finish( callback )
            return
        tempFile.seek( 0 )
        data = tempFile.read()
        self.set_header( 'Content-type', 'image/png' )
        self.set_header( 'Content-Disposition', 'attachment; filename="%s.png"' % md5( plotImageFile ).hexdigest() )
        self.set_header( 'Content-Length', len( data ) )
        self.set_header( 'Content-Transfer-Encoding', 'Binary' )
    #self.set_header( 'Cache-Control', "no-cache, no-store, must-revalidate, max-age=0" )
    #self.set_header( 'Pragma', "no-cache" )
    #self.set_header( 'Expires', ( datetime.datetime.utcnow() - datetime.timedelta( minutes = -10 ) ).strftime( "%d %b %Y %H:%M:%S GMT" ) )
        self.finish( data )



    def __parseFormParams(self):
        params = self.request.arguments

        self.log.always(params)
        pD = {}
        extraParams = {}
        for name in params:
            if name =='_Source':
                value = ",".join(params['_Source'])
                pD['Source']=str( value )
            if name =='_Destination':
                value = ",".join(params['_Destination'])
                pD['Destination']=str( value )
            if name =='_User':
                value = ",".join(params['_User'])
                pD['User']=str( value )
        if('_Source' in  params.keys()):
            del(params['_Source'])
        if('_Destination' in  params.keys()):
            del(params['_Destination'])
        if('_User' in  params.keys()):
            del(params['_User'])

        for name in params:
            if name.find( "_" ) != 0:
              continue
            value = params[ name ][0]
            name = name[1:]
            pD[ name ] = str( value )
        self.log.always(pD)

        # Get plotname
        if not 'grouping' in pD:
           return S_ERROR( "Missing grouping!" )
        grouping = pD[ 'grouping' ]
        # Get plotname
        if not 'typeName' in pD:
           return S_ERROR( "Missing type name!" )
        typeName = pD[ 'typeName' ]
        del( pD[ 'typeName' ] )
        # Get plotname
        if not 'plotName' in pD:
           return S_ERROR( "Missing plot name!" )
        reportName = pD[ 'plotName' ]
        del( pD[ 'plotName' ] )
        # Get times
        start =datetime.datetime.strptime(pD[ 'startTime' ],"%Y-%m-%d %H:%M")
        end = datetime.datetime.strptime(pD[ 'endTime' ],"%Y-%m-%d %H:%M")
        del( pD[ 'startTime' ] )
        del( pD[ 'endTime' ] )
        # Get extraParams

        # Listify the rest
        for selName in pD:
          pD[ selName ] = List.fromChar( pD[ selName ], "," )
          self.log.always("11111111111111111111111111111111111111111111111111111111111111111111111111111")
          self.log.always(pD[ selName ])
        return S_OK( ( typeName, reportName, start, end, pD, grouping, extraParams ) )



#

import json

from DIRAC                import gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Base.DB   import DB


__RCSID__ = "$Id: TaskDB.py 1 2015-01-11 11:39:29 zhaoxh@ihep.ac.cn $"

class TaskDB( DB ):

  def __init__( self, maxQueueSize = 10 ):
    DB.__init__( self, 'TaskDB', 'WorkloadManagement/TaskDB', maxQueueSize )

    result = self.__initializeDB()
    if not result[ 'OK' ]:
      raise Exception( "Can't create tables: %s" % result[ 'Message' ] )

  def __initializeDB( self ):
    """
    Create the tables
    """
    result = self._query( "show tables" )
    if not result[ 'OK' ]:
      return result

    tablesInDB = [ t[0] for t in result[ 'Value' ] ]
    tablesToCreate = {}
    self.__tablesDesc = {}

    self.__tablesDesc[ 'Task' ] = { 'Fields' : { 'TaskID'       : 'BIGINT UNSIGNED AUTO_INCREMENT NOT NULL',
                                                 'TaskName'     : 'VARCHAR(128) NOT NULL DEFAULT "unknown"',
                                                 'CreationTime' : 'DATETIME NOT NULL',
                                                 'UpdateTime'   : 'DATETIME NOT NULL',
                                                 'Status'       : 'VARCHAR(64) NOT NULL DEFAULT "Unknown"',
                                                 'Owner'        : 'VARCHAR(64) NOT NULL DEFAULT "unknown"',
                                                 'OwnerDN'      : 'VARCHAR(255) NOT NULL DEFAULT "unknown"',
                                                 'OwnerGroup'   : 'VARCHAR(128) NOT NULL DEFAULT "unknown"',
                                                 'Site'         : 'VARCHAR(512) NOT NULL DEFAULT ""',
                                                 'JobGroup'     : 'VARCHAR(512) NOT NULL DEFAULT ""',
                                                 'Progress'     : 'VARCHAR(128) NOT NULL DEFAULT "{}"',
                                                 'Info'         : 'TEXT NOT NULL',
                                               },
                                    'PrimaryKey' : 'TaskID',
                                    'Indexes': { 'TaskIDIndex'       : [ 'TaskID' ],
                                                 'CreationTimeIndex' : [ 'CreationTime' ],
                                                 'UpdateTimeIndex'   : [ 'UpdateTime' ],
                                                 'StatusIndex'       : [ 'Status' ],
                                                 'OwnerIndex'        : [ 'Owner' ],
                                                 'OwnerDNIndex'      : [ 'OwnerDN' ],
                                                 'OwnerGroupIndex'   : [ 'OwnerGroup' ],
                                               }
                                  }

    self.__tablesDesc[ 'TaskHistory' ] = { 'Fields' : { 'TaskHistoryID' : 'BIGINT UNSIGNED AUTO_INCREMENT NOT NULL',
                                                        'TaskID'        : 'BIGINT UNSIGNED NOT NULL DEFAULT 0',
                                                        'Status'        : 'VARCHAR(64) NOT NULL DEFAULT "Unknown"',
                                                        'StatusTime'    : 'DATETIME NOT NULL',
                                                        'Description'   : 'VARCHAR(128) NOT NULL DEFAULT ""',
                                                      },
                                           'PrimaryKey' : 'TaskHistoryID',
                                           'Indexes': { 'TaskHistoryIDIndex' : [ 'TaskHistoryID' ],
                                                        'TaskIDIndex'        : [ 'TaskID' ],
                                                      }
                                         }

    self.__tablesDesc[ 'TaskJob' ] = { 'Fields' : { 'TaskJobID'    : 'BIGINT UNSIGNED AUTO_INCREMENT NOT NULL',
                                                    'TaskID'       : 'BIGINT UNSIGNED NOT NULL DEFAULT 0',
                                                    'JobID'        : 'BIGINT UNSIGNED UNIQUE NOT NULL DEFAULT 0',
                                                    'Info'         : 'TEXT NOT NULL',
                                                  },
                                       'PrimaryKey' : 'TaskJobID',
                                       'Indexes': { 'TaskJobIDIndex' : [ 'TaskJobID' ],
                                                    'TaskIDIndex'    : [ 'TaskID' ],
                                                    'JobIDIndex'     : [ 'JobID' ],
                                                  }
                                     }

    for tableName in self.__tablesDesc:
      if not tableName in tablesInDB:
        tablesToCreate[ tableName ] = self.__tablesDesc[ tableName ]

    return self._createTables( tablesToCreate )

################################################################################

  def createTask( self, taskName, status, owner, ownerDN, ownerGroup, taskInfo = {} ):
    taskAttrNames = ['TaskName', 'CreationTime', 'UpdateTime', 'Status', 'Owner', 'OwnerDN', 'OwnerGroup', 'Info']
    taskAttrValues = [taskName, Time.dateTime(), Time.dateTime(), status, owner, ownerDN, ownerGroup, json.dumps(taskInfo, separators=(',',':'))]

    result = self.insertFields( 'Task', taskAttrNames, taskAttrValues )
    if not result['OK']:
      self.log.error( 'Can not create new task', result['Message'] )
      return result

    if 'lastRowId' not in result:
      return S_ERROR( 'Failed to retrieve a new ID for task' )

    taskID = int( result['lastRowId'] )

    self.log.info( 'TaskDB: New TaskID served "%s"' % taskID )

    return S_OK( taskID )

  def addTaskJob( self, taskID, jobID, jobInfo ):
    taskJobAttrNames = ['TaskID', 'JobID', 'Info']
    taskJobAttrValues = [taskID, jobID, json.dumps(jobInfo, separators=(',',':'))]

    result = self.insertFields( 'TaskJob', taskJobAttrNames, taskJobAttrValues )
    if not result['OK']:
      self.log.error( 'Can not add job to task %s' % taskID, result['Message'] )

    return result

  def insertTaskHistory( self, taskID, status, description = '' ):
    taskHistoryAttrNames = ['TaskID', 'Status', 'StatusTime', 'Description']
    taskHistoryAttrValues = [taskID, status, Time.dateTime(), description]

    result = self.insertFields( 'TaskHistory', taskHistoryAttrNames, taskHistoryAttrValues )
    if not result['OK']:
      self.log.error( 'Can not insert task history to task %s' % taskID, result['Message'] )

    return result

  def updateTask( self, taskID, taskAttrNames, taskAttrValues ):
    condDict = { 'TaskID': taskID }

    if 'UpdateTime' not in taskAttrNames:
      taskAttrNames.append( 'UpdateTime' )
      taskAttrValues.append( Time.dateTime() )

    result = self.updateFields( 'Task', taskAttrNames, taskAttrValues, condDict )
    if not result['OK']:
      self.log.error( 'Can not update task %s' % taskID, result['Message'] )

    return result

  def updateTaskStatus( self, taskID, status, description ):
    result = self.updateTask( taskID, ['Status'], [status] )
    if not result['OK']:
      return result
    result = self.insertTaskHistory( taskID, status, description )
    if not result['OK']:
      return result

    return S_OK( status )

  def updateTaskProgress( self, taskID, progress ):
    condDict = { 'TaskID': taskID }
    taskAttrNames = ['Progress']
    taskAttrValues = [json.dumps(progress, separators=(',',':'))]

    result = self.updateFields( 'Task', taskAttrNames, taskAttrValues, condDict )
    if not result['OK']:
      self.log.error( 'Can not update task progress for task %s' % taskID, result['Message'] )

    return result

  def updateTaskInfo( self, taskID, taskInfo ):
    condDict = { 'TaskID': taskID }
    taskAttrNames = ['Info']
    taskAttrValues = [json.dumps(taskInfo, separators=(',',':'))]

    result = self.updateFields( 'Task', taskAttrNames, taskAttrValues, condDict )
    if not result['OK']:
      self.log.error( 'Can not update task info for task %s' % taskID, result['Message'] )

    return result

  def renameTask( self, taskID, newName ):
    condDict = { 'TaskID': taskID }
    taskAttrNames = ['TaskName']
    taskAttrValues = [newName]

    result = self.updateFields( 'Task', taskAttrNames, taskAttrValues, condDict )
    if not result['OK']:
      self.log.error( 'Can not rename task %s' % taskID, result['Message'] )

    return result

################################################################################

  def getDistinctTaskAttributes( self, attribute, condDict = None, older = None,
                                newer = None, timeStamp = 'UpdateTime' ):
    """ Get distinct values of the task attribute under specified conditions
    """
    return self.getDistinctAttributeValues( 'Task', attribute, condDict = condDict,
                                              older = older, newer = newer, timeStamp = timeStamp )

  def getTaskCount( self, condDict ):
    newer = None
    if 'FromDate' in condDict:
      newer = condDict['FromDate']
      del condDict['FromDate']
    older = None
    if 'ToDate' in condDict:
      older = condDict['ToDate']
      del condDict['ToDate']

    result = self.getCounters( 'Task', ['Status'], condDict, newer = newer, older = older, timeStamp = 'UpdateTime' )
    if not result['OK']:
      self.log.error( 'Can not get task count', result['Message'] )
      return result

    nTasks = 0
    for stDict, count in result['Value']:
      nTasks += count

    return S_OK( nTasks )

  def getTasks( self, outFields, condDict, limit = None, offset = None, orderAttribute = None ):
    newer = None
    if 'FromDate' in condDict:
      newer = condDict['FromDate']
      del condDict['FromDate']
    older = None
    if 'ToDate' in condDict:
      older = condDict['ToDate']
      del condDict['ToDate']

    result = self.getFields( 'Task', outFields, condDict, newer = newer, older = older, timeStamp = 'UpdateTime', limit = (limit, offset), orderAttribute = orderAttribute )
    if not result['OK']:
      self.log.error( 'Can not get task list', result['Message'] )
      return result

    return S_OK( result['Value'] )

  def getAttributesForTaskList( self, taskIDs, attrList = None ):
    """ Get attributes for the jobs in the the jobIDList.
        Returns an S_OK structure with a dictionary of dictionaries as its Value:
        ValueDict[jobID][attribute_name] = attribute_value
    """
    if not taskIDs:
      return S_OK( {} )
    if attrList:
      attrNames = ',' + ','.join( [ str( x ) for x in attrList ] )
      attr_tmp_list = attrList
    else:
      attrNames = ''
      attr_tmp_list = []
    taskList = ','.join( [str( x ) for x in taskIDs] )

    # FIXME: need to check if the attributes are in the list of task Attributes

    cmd = 'SELECT TaskID%s FROM Task WHERE TaskID in ( %s )' % ( attrNames, taskList )
    res = self._query( cmd )
    if not res['OK']:
      return res
    try:
      retDict = {}
      for retValues in res['Value']:
        taskID = retValues[0]
        taskDict = {}
        taskDict[ 'TaskID' ] = taskID
        attrValues = retValues[1:]
        for i in range( len( attr_tmp_list ) ):
          try:
            taskDict[attr_tmp_list[i]] = attrValues[i].tostring()
          except Exception:
            taskDict[attr_tmp_list[i]] = str( attrValues[i] )
        retDict[int( taskID )] = taskDict
      return S_OK( retDict )
    except Exception, x:
      return S_ERROR( 'TaskDB.getAttributesForTaskList: Failed\n%s' % str( x ) )

  def getTask( self, taskID, outFields ):
    condDict = { 'TaskID': taskID }
    result = self.getFields( 'Task', outFields, condDict )
    if not result['OK']:
      self.log.error( 'Can not get task %s' % taskID, result['Message'] )
      return result

    if not result['Value']:
      self.log.error( 'Task ID %d not found' % taskID )
      return S_ERROR( 'Task ID %d not found' % taskID )

    return S_OK( result['Value'][0] )

  def getTaskStatus( self, taskID ):
    outFields = ( 'Status', )
    result = self.getTask( taskID, outFields )
    if not result['OK']:
      self.log.error( 'Can not get task status for task %s' % taskID, result['Message'] )
      return result

    return S_OK( result['Value'][0] )

  def getTaskInfo( self, taskID ):
    outFields = ( 'Info', )
    result = self.getTask( taskID, outFields )
    if not result['OK']:
      self.log.error( 'Can not get task info for task %s' % taskID, result['Message'] )
      return result

    return S_OK( json.loads( result['Value'][0] ) )

  def getTaskHistories( self, taskID ):
    condDict = { 'TaskID': taskID }
    outFields = ( 'Status', 'StatusTime', 'Description' )
    result = self.getFields( 'TaskHistory', outFields, condDict )
    if not result['OK']:
      self.log.error( 'Can not get task histories for task %s' % taskID, result['Message'] )
      return result

    return S_OK( result['Value'] )

  def getTaskJobs( self, taskID ):
    condDict = { 'TaskID': taskID }
    outFields = ( 'JobID', )
    result = self.getFields( 'TaskJob', outFields, condDict, orderAttribute = 'JobID:ASC' )
    if not result['OK']:
      self.log.error( 'Can not get task jobs for task %s' % taskID, result['Message'] )
      return result

    return S_OK( [ i[0] for i in  result['Value'] ] )

  def getJobs( self, jobIDs, outFields ):
    if not jobIDs:
      return S_OK( [] )

    condDict = { 'JobID': jobIDs }
    result = self.getFields( 'TaskJob', outFields, condDict )
    if not result['OK']:
      self.log.error( 'Can not get task ID for job ID %s' % jobIDs, result['Message'] )
      return result

    return S_OK( result['Value'] )

  def getTaskIDFromJob( self, jobID ):
    condDict = { 'JobID': jobID }
    outFields = ( 'TaskID', )
    result = self.getFields( 'TaskJob', outFields, condDict )
    if not result['OK']:
      self.log.error( 'Can not get task ID for job %s' % jobID, result['Message'] )
      return result

    return S_OK( [ i[0] for i in  result['Value'] ] )

  def getJobInfo( self, jobID ):
    condDict = { 'JobID': jobID }
    outFields = ( 'Info', )
    result = self.getFields( 'TaskJob', outFields, condDict )
    if not result['OK']:
      self.log.error( 'Can not get job info for job %s' % jobID, result['Message'] )
      return result

    if not result['Value']:
      self.log.error( 'Job info %d not found' % jobID )
      return S_ERROR( 'Job info %d not found' % jobID )

    return S_OK( json.loads( result['Value'][0][0] ) )



""" JobIHEPCommand

  The JobIHEPCommand class is a command class to know about present jobs efficiency

"""

from DIRAC                                                         import S_OK, S_ERROR
from DIRAC.Core.DISET.RPCClient                                    import RPCClient
from DIRAC.ResourceStatusSystem.Command.Command                    import Command
from IHEPDIRAC.ResourceStatusSystem.Utilities                      import CSHelpers
from IHEPDIRAC.ResourceStatusSystem.Client.ResourceManagementIHEPClient import ResourceManagementIHEPClient
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

__RCSID__ = '$Id:  $'

class JobIHEPCommand( Command ):
  """
    Job "master" Command.
  """

  def __init__( self, args = None, clients = None ):

    super( JobIHEPCommand, self ).__init__( args, clients )

    if 'WMSAdministrator' in self.apis:
      self.wmsAdmin = self.apis[ 'WMSAdministrator' ]
    else:
      self.wmsAdmin = RPCClient( 'WorkloadManagement/WMSAdministrator' )

    if 'ResourceManagementIHEPClient' in self.apis:
      self.rmIHEPClient = self.apis[ 'ResourceManagementIHEPClient' ]
    else:
      self.rmIHEPClient = ResourceManagementIHEPClient()

  def _storeCommand( self, result ):
    """
      Stores the results of doNew method on the database.
    """

    for jobDict in result:

      resQuery = self.rmIHEPClient.addOrModifyJobCache( jobDict[ 'Site' ],
                                                    jobDict[ 'MaskStatus' ],
                                                    jobDict[ 'Efficiency' ],
                                                    jobDict[ 'Running' ],
                                                    jobDict[ 'Waiting' ],
                                                    jobDict[ 'Done' ],
                                                    jobDict[ 'Failed' ],
                                                    jobDict[ 'Completed' ],
                                                    jobDict[ 'Stalled' ],
                                                    jobDict[ 'Status' ])
      if not resQuery[ 'OK' ]:
        return resQuery
    return S_OK()

  def _prepareCommand( self ):
    """
      JobIHEPCommand requires one arguments:
      - name : <str>
    """

    if not 'name' in self.args:
      return S_ERROR( '"name" not found in self.args' )
    name = self.args[ 'name' ]

    return S_OK( name )

  def doNew( self, masterParams = None ):
    """
      Gets the parameters to run, either from the master method or from its
      own arguments.

      It contacts the WMSAdministrator with a list of site names, or a single
      site.

      If there are jobs, are recorded and then returned.
    """

    if masterParams is not None:
      name = masterParams
    else:
      params = self._prepareCommand()
      if not params[ 'OK' ]:
        return params
      name = params[ 'Value' ]

#    resultMask = self.wmsAdmin.getSiteMask()
    diracAdmin = DiracAdmin()
    resultMask = diracAdmin.getSiteMask()
    if not resultMask[ 'OK' ]:
      return resultMask
    resultMask = resultMask[ 'Value' ]

    # selectDict, sortList, startItem, maxItems
    # Returns statistics of Last day !
    results = self.wmsAdmin.getSiteSummaryWeb( { 'Site' : name }, [], 0, 0 )
    if not results[ 'OK' ]:
      return results
    results = results[ 'Value' ]

    if not 'ParameterNames' in results:
      return S_ERROR( 'Wrong result dictionary, missing "ParameterNames"' )
    params = results[ 'ParameterNames' ]

    if not 'Records' in results:
      return S_ERROR( 'Wrong formed result dictionary, missing "Records"' )
    records = results[ 'Records' ]

    uniformResult = []
    siteJobs = {}

    for record in records:

      # This returns a dictionary with the following keys
      # 'Site', 'GridType', 'Country', 'Tier', 'MaskStatus', 'Received',
      # 'Checking', 'Staging', 'Waiting', 'Matched', 'Running', 'Stalled',
      # 'Done', 'Completed', 'Failed', 'Efficiency', 'Status'
      jobDict = dict( zip( params , record ))
      siteJobs[ jobDict.pop( 'Site' ) ] = jobDict

#       # We cast efficiency to a float
#       jobDict[ 'Efficiency' ] = float( jobDict[ 'Efficiency' ] )

#       uniformResult.append( jobDict )

    for site in name:
      recordDict = {}
      recordDict[ 'Site' ] = site
      if site in resultMask:
        recordDict[ 'MaskStatus' ] = 'Active'
      else:
        recordDict[ 'MaskStatus' ] = 'Banned'

      if siteJobs.has_key( site ):
#        recordDict[ 'MaskStatus' ] = siteJobs[ site ][ 'MaskStatus' ]
        recordDict[ 'Running' ] = siteJobs[ site ][ 'Running' ]
        recordDict[ 'Waiting' ] = siteJobs[ site ][ 'Waiting' ] + siteJobs[ site ][ 'Checking' ]
        recordDict[ 'Done' ] = siteJobs[ site ][ 'Done' ]
        recordDict[ 'Failed' ] = siteJobs[ site ][ 'Failed' ]
        recordDict[ 'Completed' ] = siteJobs[ site ][ 'Completed' ]
        recordDict[ 'Stalled' ] = siteJobs[ site ][ 'Stalled' ]
        recordDict[ 'Efficiency' ] = float( siteJobs[ site ][ 'Efficiency' ] )
        recordDict[ 'Status' ] = siteJobs[ site ][ 'Status' ]
      else:
        recordDict[ 'Running' ] = 0
        recordDict[ 'Waiting' ] = 0
        recordDict[ 'Done' ] = 0
        recordDict[ 'Failed' ] = 0
        recordDict[ 'Completed' ] = 0
        recordDict[ 'Stalled' ] = 0
        recordDict[ 'Efficiency' ] = 0.0
        recordDict[ 'Status' ] = 'Idle'

      uniformResult.append( recordDict )

    storeRes = self._storeCommand( uniformResult )
    if not storeRes[ 'OK' ]:
      return storeRes

    return S_OK( uniformResult )

  def doCache( self ):
    """
      Method that reads the cache table and tries to read from it. It will
      return a list of dictionaries if there are results.
    """

    params = self._prepareCommand()
    if not params[ 'OK' ]:
      return params
    name = params[ 'Value' ]

    result = self.rmIHEPClient.selectJobCache( name )
    if result[ 'OK' ]:
      result = S_OK( [ dict( zip( result[ 'Columns' ], res ) ) for res in result[ 'Value' ] ] )

    return result

  def doMaster( self ):
    """
      Master method.

      Gets all sites and calls doNew method.
    """

    siteNames = CSHelpers.getSites()
    if not siteNames[ 'OK' ]:
      return siteNames
    siteNames = siteNames[ 'Value' ]

    jobsResults = self.doNew( siteNames )
    if not jobsResults[ 'OK' ]:
      self.metrics[ 'failed' ].append( jobsResults[ 'Message' ] )

    return S_OK( self.metrics )

################################################################################
#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



import os

from DIRAC import gConfig, gLogger

class LocalMount(object):
  def __init__(self):
    super(LocalMount, self).__init__()
    self._directlyRead = True

    self.__mountPoint = gConfig.getValue('/Resources/Applications/DataLocation/LocalMount/Prefix', '/dcache/bes')
    gLogger.debug('Mount point:', self.__mountPoint)

  def _available(self):
    return os.path.isdir(self.__mountPoint)

  def _lfnToRemote(self, lfn):
    return self.__mountPoint + lfn



import os
import time

import subprocess

from DIRAC import gLogger

class MergeFile(object):
  def __init__(self):
    self._directlyRead = False
    self._localValidation = True

  def merge(self, fileList, outputDir, mergeName, mergeExt, mergeMaxSize, mergeCallback):
    allFileSize = self.__getAllFileSize(fileList)

    count = 0
    tempSize = 0
    tempList = []
    for i in range(len(fileList)):
      fn = fileList[i]
      tempSize += allFileSize[fn]
      tempList.append(fn)
      if i == len(fileList) - 1 or tempSize > mergeMaxSize or tempSize + allFileSize[fileList[i+1]] > mergeMaxSize:
        count += 1
        mergePath = os.path.join(outputDir, '%s_%04d%s' % (mergeName, count, mergeExt))

        startTime = time.time()
        ret = self.__doMerge(tempList, mergePath)
        endTime = time.time()

        if mergeCallback is not None:
          mergeCallback(tempList, mergePath, tempSize, endTime-startTime, ret)
        if not ret:
          return False

        tempSize = 0
        tempList = []

    return True

  def __doMerge(self, fileList, mergePath):
    if len(fileList) == 0:
      gLogger.error('Can not merge empty file list!')
      return False

    try:
      args = ['hadd', mergePath] + fileList
      p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      out, err = p.communicate()
      ret = p.returncode
    except Exception, e:
      gLogger.debug('hadd error:', e)
      gLogger.error('Command "hadd" not found. Can not merge files. Please check your environment!')
      return False

    return ret == 0

  def __getAllFileSize(self, fileList):
    allFileSize = {}
    for fn in fileList:
      allFileSize[fn] = os.path.getsize(fn)
    return allFileSize



''' EmailAction

  This action writes all the necessary data to a cache file ( cache.db ) that
  will be used later by the EmailAgent in order to send the emails for each site.

'''

import os
import sqlite3
from DIRAC                                                      import gConfig, S_ERROR, S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.Actions.BaseAction import BaseAction
from DIRAC.Core.Utilities.SiteSEMapping                         import getSitesForSE

__RCSID__ = '$Id:  $'

class EmailAction( BaseAction ):

  def __init__( self, name, decisionParams, enforcementResult, singlePolicyResults,
                clients = None ):

    super( EmailAction, self ).__init__( name, decisionParams, enforcementResult,
                                         singlePolicyResults, clients )

    if 'DIRAC' in os.environ:
      self.cacheFile = os.path.join( os.getenv('DIRAC'), 'work/ResourceStatus/cache.db' )
    else:
      self.cacheFile = os.path.realpath('cache.db')

  def run( self ):
    ''' Checks it has the parameters it needs and writes the date to a cache file.
    '''
    # Minor security checks

    element = self.decisionParams[ 'element' ]
    if element is None:
      return S_ERROR( 'element should not be None' )

    name = self.decisionParams[ 'name' ]
    if name is None:
      return S_ERROR( 'name should not be None' )

    statusType = self.decisionParams[ 'statusType' ]
    if statusType is None:
      return S_ERROR( 'statusType should not be None' )

    previousStatus = self.decisionParams[ 'status' ]
    if previousStatus is None:
      return S_ERROR( 'status should not be None' )

    status = self.enforcementResult[ 'Status' ]
    if status is None:
      return S_ERROR( 'status should not be None' )

    reason = self.enforcementResult[ 'Reason' ]
    if reason is None:
      return S_ERROR( 'reason should not be None' )

    siteName = getSitesForSE(name)

    if not siteName['OK']:
      self.log.error('Resource %s does not exist at any site: %s' % (name, siteName['Message']))
      siteName = "Unassigned Resources"
    elif not siteName['Value']:
      siteName = "Unassigned Resources"
    else:
      siteName = siteName['Value'][0]

    if status != previousStatus:
      with sqlite3.connect(self.cacheFile) as conn:

        try:
          conn.execute('''CREATE TABLE IF NOT EXISTS ResourceStatusCache(
                        SiteName VARCHAR(64) NOT NULL,
                        ResourceName VARCHAR(64) NOT NULL,
                        Status VARCHAR(8) NOT NULL DEFAULT "",
                        PreviousStatus VARCHAR(8) NOT NULL DEFAULT "",
                        StatusType VARCHAR(128) NOT NULL DEFAULT "all",
                        Time TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                       );''')

        except sqlite3.OperationalError:
          self.log.error('Email cache database is locked')

        conn.execute("INSERT INTO ResourceStatusCache (SiteName, ResourceName, Status, PreviousStatus, StatusType)"
                     " VALUES ('" + siteName + "', '" + name + "', '" + status + "', '" + previousStatus + "', '" + statusType + "' ); "
                    )

        conn.commit()

    return S_OK()

################################################################################
#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



#!/usr/bin/env python

import hashlib

class Md5CheckSum():
  @staticmethod
  def checksum(filename, blocksize=1048576):
    hash = hashlib.md5()
    with open(filename, "rb") as f:
      for block in iter(lambda: f.read(blocksize), ""):
        hash.update(block)
    return hash.hexdigest()



extra_dErrName = {
        'EAPPCOMMON'       : 1,
        'EAPPCMDNOTFOUND'  : 127,
        'EAPPSIGABRT'      : 134,
        'EAPPKILLED'       : 137,

        'EBOSSRANTRGERROR'     : 66,
        'EBOSSCVMFS'           : 70,
        'EBOSSTOOMANYRUNS'     : 71,
        'EBOSSUPLOAD'          : 72,

        'ECEPCCVMFS'           : 11,
        'ECEPCDOWNLOADINPUT'   : 12,
        'ECEPCSIM'             : 20,
        'ECEPCDATABASE'        : 21,
        'ECEPCTOOMANYSUBSTEPS' : 22,
        'ECEPCSIMNOEVENTS'     : 23,
        'ECEPCREC'             : 30,
        'ECEPCRECNOEVENTS'     : 31,
        'ECEPCUPLOAD'          : 51,
}

extra_dErrorCode = {
        1   : 'EAPPCOMMON',
        127 : 'EAPPCMDNOTFOUND',
        134 : 'EAPPSIGABRT',
        137 : 'EAPPKILLED',

        66 : 'EBOSSRANTRGERROR',
        70 : 'EBOSSCVMFS',
        71 : 'EBOSSTOOMANYRUNS',
        72 : 'EBOSSUPLOAD',

        11 : 'ECEPCCVMFS',
        12 : 'ECEPCDOWNLOADINPUT',
        20 : 'ECEPCSIM',
        21 : 'ECEPCDATABASE',
        22 : 'ECEPCTOOMANYSUBSTEPS',
        23 : 'ECEPCSIMNOEVENTS',
        30 : 'ECEPCREC',
        31 : 'ECEPCRECNOEVENTS',
        51 : 'ECEPCUPLOAD',
}

extra_dStrError = {
        1   : 'Common exit error',
        127 : 'Command or library not found',
        134 : 'Aborted with SIGABRT',
        137 : 'Killed by system',

        66 : 'BOSS random trigger download error',
        70 : 'BOSS CVMFS not found',
        71 : 'BOSS too many runs for reconstruction',
        72 : 'BOSS upload output data error',

        11 : 'CEPC CVMFS not found',
        12 : 'CEPC download input error',
        20 : 'CEPC simulation error',
        21 : 'CEPC database connection error',
        22 : 'CEPC too many substeps',
        23 : 'CEPC simulation insufficient events',
        30 : 'CEPC reconstruction error',
        31 : 'CEPC reconstruction insufficient events',
        51 : 'CEPC upload output data error',
}



import json, math
from DIRAC                      import S_OK
from WebAppDIRAC.Lib.WebHandler import WebHandler, WErr, WOK, asyncGen
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin


class SiteStatusMonitorHandler( WebHandler ):

  AUTH_PROPS = "authenticated"

  @asyncGen
  def web_getSelectionData( self ):
    publisher = RPCClient( 'ResourceStatus/PublisherIHEP' )

    callback = {}
    ret = { 'success' : 'true', 'result' : callback }

    data = self.getSessionData()
    vo = yield self.threadTask( publisher.getVOByGroup, data[ 'user' ][ 'group' ] )
    vo = vo[ 'Value' ]
    if vo != '':
      ret[ 'defaultVO' ] = vo

    allSites = yield self.threadTask( publisher.getSites )
    if allSites[ 'OK' ]:
      if vo != '':
        sites = []
        for site in allSites[ 'Value' ]:
          if vo in publisher.getSiteVO( site )[ 'Value' ]:
            sites.append( [ site ] )
      else:
        sites = [ [ site ] for site in allSites[ 'Value' ] ]
    else:
      sites = [ [ 'Error happened on service side' ] ]

    types = yield self.threadTask( publisher.getDomains )
    if types[ 'OK' ]:
      types = [ [ type ] for type in types[ 'Value' ] ]
    else:
      types = [ [ 'Error happened on service side' ] ]

    mask = [ [ 'Active' ], [ 'Banned' ] ]

    vos = yield self.threadTask( publisher.getVOs )
    if vos[ 'OK' ]:
      vos = [ [ vo ] for vo in vos[ 'Value' ] ]
    else:
      vos = [ [ 'Error happened on service side' ] ]

    callback[ 'site' ] = sites
    callback[ 'type' ] = types
    callback[ 'mask' ] = mask
    callback[ 'vo' ] = vos

    self.finish( ret )


  @asyncGen
  def web_getMainData( self ):
    publisher = RPCClient( 'ResourceStatus/PublisherIHEP' )

    sitesDict = {}

    sites = yield self.threadTask( publisher.getSites )
    if not sites[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : sites[ 'Message' ] } )
      return
    sites = sites[ 'Value' ]

    sitesType = self.__getSitesType( sites )

    sitesMask = yield self.threadTask( self.__getSitesMaskStatus, sites )
    if not sitesMask[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : sitesMask[ 'Message' ] } )
      return
    sitesMask = sitesMask[ 'Value' ]

    sitesVO= yield self.threadTask( self.__getSitesVO, sites )
    if not sitesVO[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : sitesVO[ 'Message' ] } )
      return
    sitesVO = sitesVO[ 'Value' ]

    for site in sites:
      sitesDict[ site ]= {}
      sitesDict[ site ][ 'SiteType' ] = sitesType[ site ]
      sitesDict[ site ][ 'MaskStatus' ] = sitesMask[ site ]
      sitesDict[ site ][ 'VO' ] = sitesVO[ site ]

    req = self._request()

    selectedSites = req.get( 'Site' )
    selectedType = req.get( 'SiteType' )
    selectedMask = req.get( 'MaskStatus' )
    selectedVOs = req.get( 'VO' )

    sitesDict = yield self.threadTask( self.__siteFilte, sitesDict, selectedSites, selectedType, selectedMask, selectedVOs )

    if not sitesDict:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : 'There are no data to display' } )
      return

    summarys = []

    samSummary = yield self.threadTask( publisher.getSitesSAMSummary, sitesDict.keys(), selectedVOs )
    if not samSummary[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : samSummary[ 'Message' ] } )
      return 
    summarys.append( samSummary[ 'Value' ] )

    storageSummary = yield self.threadTask( publisher.getSitesStorageSummary, sitesDict.keys() )
    if not storageSummary[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : storageSummary[ 'Message' ] } )
      return 
    summarys.append( storageSummary[ 'Value' ] )

    jobSummary = yield self.threadTask( publisher.getSitesJobSummary, sitesDict.keys() )
    if not jobSummary[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : jobSummary[ 'Message' ] } )
      return 
    summarys.append( jobSummary[ 'Value' ] )

    wnSummary = yield self.threadTask( publisher.getSitesWNSummary, sitesDict.keys() )
    if not wnSummary[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : wnSummary[ 'Message' ] } )
      return 
    summarys.append( wnSummary[ 'Value' ] )

    for summary in summarys:
      for site, summaryDict in summary.items():
        sitesDict[ site ].update( summaryDict )

    callback = []

    for siteName, siteDict in sitesDict.items():
      siteDict[ 'Site' ] = siteName
      callback.append( siteDict )

    self.finish( { 'success' : 'true', 'result' : callback, 'total' : len( callback ) } )


  @asyncGen        
  def web_getSAMData( self ):
    publisher = RPCClient( 'ResourceStatus/PublisherIHEP' )

    req = self._request()
    site = req[ 'Site' ][ -1 ].encode()
    if req.has_key( 'VO' ):
      vo = req[ 'VO' ][ -1 ].encode()
    else:
      vo = 'all'

    samSummary = yield self.threadTask( publisher.getSAMSummary, site, vo )
    if not samSummary[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : samSummary[ 'Message' ] } )
      return
    samSummary = samSummary[ 'Value' ]

    callback = []
    columns = []
    for elementName, samDict in samSummary.items():
      columns += samDict.keys()
      samDict[ 'ElementName' ] = elementName
      callback.append( samDict )
    columns = list( set( columns ) )
    columns.remove( 'ElementType' )
    columns.remove( 'Status' )

    self.finish( { 'success' : 'ture', 'result' : callback, 'columns' : columns, 'total' : len( callback ) } )


  @asyncGen
  def web_getSAMDetail( self ):
    publisher = RPCClient( 'ResourceStatus/PublisherIHEP' )

    elementName = self.request.arguments[ 'elementName' ][ -1 ]
    testType = self.request.arguments[ 'testType' ][ -1 ]

    samDetail = yield self.threadTask( publisher.getSAMDetail, elementName, testType )
    if not samDetail[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : samDetail[ 'Message' ] } )
      return
    samDetail = samDetail[ 'Value' ]

    samDetail[ 'SubmissionTime' ] = samDetail[ 'SubmissionTime' ].strftime( '%Y-%m-%d %H:%M' ) + ' UTC'
    if samDetail[ 'CompletionTime' ]:
      samDetail[ 'CompletionTime' ] = samDetail[ 'CompletionTime' ].strftime( '%Y-%m-%d %H:%M' ) + ' UTC'
    if samDetail[ 'ApplicationTime' ]:
      samDetail[ 'ApplicationTime' ] = str( samDetail[ 'ApplicationTime' ] ) + ' seconds'
    for key in samDetail:
      samDetail[ key ] = samDetail[ key ] or '-'

    self.finish( { 'success' : 'ture', 'result' : samDetail } )


  @asyncGen        
  def web_getWorkNodeData( self ):
    publisher = RPCClient( 'ResourceStatus/PublisherIHEP' )

    req = self._request()
    site = req[ 'Site' ][ -1 ].encode()

    wnsInfo = yield self.threadTask( publisher.getSiteWNsInfo, site )
    if not wnsInfo[ 'OK' ]:
      self.finish( { 'success' : 'false', 'result' : [], 'total' : 0, 'error' : wnsInfo[ 'Message' ] } )
      return
    wnsInfo = wnsInfo[ 'Value' ]

    totalRunning = 0
    totalDone = 0
    totalFailed = 0
    for record in wnsInfo:
      totalDone += record[ 'Done' ]
      totalFailed += record[ 'Failed' ]  
    if totalDone == 0 and totalFailed == 0:
      efficiency = ''
    else:
      efficiency = math.floor(float(totalDone) / (totalDone + totalFailed) * 1000) /10

    wnsInfo.append({ 'Host' : 'total',
                    'Done' : totalDone,
                    'Failed' : totalFailed,
                    'Efficiency' : efficiency })

    self.finish( { 'success' : 'ture', 'result' : wnsInfo, 'total' : len( wnsInfo ) } )


  def __siteFilte( self, sitesDict, selectedSites = None, selectedType = None, selectedMask = None, selectedVOs = None ):
    for siteName, siteDict in sitesDict.items():

      if selectedSites is not None and siteName not in selectedSites:
        del sitesDict[ siteName ]
        continue

      siteType = siteDict[ 'SiteType' ]
      if selectedType is not None and siteType not in selectedType:
        del sitesDict[ siteName ]
        continue

      siteMask = siteDict[ 'MaskStatus' ]
      if selectedMask is not None and siteMask not in selectedMask:
        del sitesDict[ siteName ]
        continue

      siteVO = siteDict[ 'VO' ]
      if selectedVOs is not None and siteVO:
        isSelected = False
        for vo in siteVO:
          if vo in selectedVOs:
            isSelected = True
            break                   
        if not isSelected:
          del sitesDict[ siteName ]

    return sitesDict


  def __getSitesType( self, sitesName ):
    sitesType = {}

    for siteName in sitesName:
      sitesType[ siteName ] = siteName.split( '.' )[ 0 ]

    return sitesType


  def __getSitesMaskStatus( self, sitesName ):
    diracAdmin = DiracAdmin()
    activeSites = diracAdmin.getSiteMask()
#    wmsAdmin = RPCClient( 'WorkloadManagement/WMSAdministrator' )
#    activeSites = wmsAdmin.getSiteMask()

    if not activeSites[ 'OK' ]:
      return activeSites
    activeSites = activeSites[ 'Value' ]

    sitesStatus = {}

    for siteName in sitesName:
      if siteName in activeSites:
        sitesStatus[ siteName ] = 'Active'
      else:
        sitesStatus[ siteName ] = 'Banned'

    return S_OK( sitesStatus )


  def __getSitesVO( self, sitesName ):
    publisher = RPCClient( 'ResourceStatus/PublisherIHEP' )

    sitesVO = {}

    for siteName in sitesName:
      vo = publisher.getSiteVO( siteName )
      if not vo[ 'OK' ]:
        return vo
      sitesVO[ siteName ] = vo[ 'Value' ]

    return S_OK( sitesVO )


  def _request(self):
    req = {}

    if 'site' in self.request.arguments:
      site = list(json.loads(self.request.arguments[ 'site' ][ -1 ]))
      if len(site) > 0:
        req[ 'Site' ] = site

    if 'type' in self.request.arguments:
      type = list(json.loads(self.request.arguments[ 'type' ][ -1 ]))
      if len(type) > 0:
        req[ 'SiteType' ] = type

    if 'mask' in self.request.arguments:
      status = list(json.loads(self.request.arguments[ 'mask' ][ -1 ]))
      if len(status) > 0:
        req[ 'MaskStatus' ] = status

    if 'vo' in self.request.arguments:
      vo = list(json.loads(self.request.arguments[ 'vo' ][ -1 ]))
      if len(vo) > 0:
        req[ 'VO' ] = vo

    return req



# -*- coding: utf-8 -*-
__RCSID__ = "$Id: $"

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule

global gTransferDB

class TransferAgent(AgentModule):

  def initialize(self):
    self.count = 0

    self.MAX_TRANSFER = self.am_getOption("MAX_TRANSFER", 2)
    self.MAX_RETRY = self.am_getOption("MAX_RETRY", 3)
    gLogger.info("MAX_TRANSFER: ", self.MAX_TRANSFER)

    self.transfer_worker = []

    global gTransferDB
    from IHEPDIRAC.TransferSystem.DB.TransferDB import TransferDB
    from IHEPDIRAC.TransferSystem.Agent.helper import helper_TransferAgent
    gTransferDB = TransferDB()

    self.helper = helper_TransferAgent(self, gTransferDB)

    return S_OK()

  def execute(self):

    gLogger.info("execute: ", self.count)
    self.count += 1

    # Handle the existed transfer worker
    for worker in self.transfer_worker[:]:
      # check worker status
      self.helper.check_worker_status(worker)
      # worker is TransferWorker
      retcode = worker.get_retcode()
      if retcode is not None:
        # Make sure the failed job can retransfer
        worker.info["retransfer"]+=1
        # Handle retcode
        worker.handle_waiting()
        result = worker.handle_exit(retcode)
        
        if result:
          gLogger.error("There is some errors!")
          gLogger.error(result)
          # if we can retransfer, try to retransfer
          if worker.info["retransfer"] < self.MAX_RETRY:
            # retransfer
            gLogger.info("Try to Retransfer (%d) "% worker.info["retransfer"])
            worker.create_popen(worker.info)
            continue
          self.helper.helper_error_report(worker, result)
          worker.info["error"] = result

        self.transfer_worker.remove(worker)
        self.helper.helper_remove_transfer(worker)
      else:
        # the job is not OK
        worker.handle_waiting()
      pass
    else:
      # the transfer work are all working
      pass

    # Create new transfer worker
    idle_worker = self.MAX_TRANSFER - len(self.transfer_worker)
    if idle_worker:
      for i in xrange(idle_worker):
        # append a new worker
        # if can't add, just break
        if not self.add_new_transfer():
          break
        gLogger.info("Add a new Transfer")

    return S_OK()

  def add_new_transfer(self):
    """
      if add a new, return true,
      else return false
    """
    # << Get New File >>
    result = self.helper.helper_get_new_request()
    gLogger.info(result)
    if not result:
      return False

    # << Add New Transfer >>
    return self.helper.helper_add_transfer(result)




""" TestExecutor

  TestExecutor is the end-point to executor SAM tests. It loads the tests which
  need to be executed dynamically and executes all the tests. At last, it stores
  the test results to database.
"""

import time
import random
import Queue
from datetime import datetime
from DIRAC                                                         import S_OK, S_ERROR, gLogger
from IHEPDIRAC.ResourceStatusSystem.Client.ResourceManagementIHEPClient import ResourceManagementIHEPClient


__RCSID__ = '$Id:  $'


class TestExecutor( object ):
  """ TestExecutor
  """

  def __init__( self, tests, apis = None ):
    """ Constructor

    examples:
      >>> tests = { 'WMS-Test' : { 'module' : 'WMSTest',
                                                           'args' : { 'executable' : [ '/usr/bin/python',
                                                                           'wms_test.py' ], 'timeout' : 1800 } } }
      >>> elements = { 'ComputingElement' : [ 'chenj01.ihep.ac.cn' ],
                                       'StorageElement', [ 'IHEPD-USER' ],
                                       'CLOUD' : [ 'CLOUD.IHEP-OPENSTACK.cn' ] }
      >>> executor = TestExecutor( tests, elements )
      >>> executor1 = TestExecutor( tests, elements, { 'ResourceManagementClient' :
                                                                                                     ResourceManagementClient() } )

    :Parameters:
      **tests** - `dict`
        dictionary with tests to be executed. The test class is loaded according to the
        'module' key and instantiated with 'args' key.
      **elements** - `dict`
        the elements need to be tested. The elements is grouped by type.
      **apis** - 'dict'
        dictionary with clients to be used in the commands issued by the policies.
        If not defined, the commands will import them.
    """

    self.apis = apis or {}
    self.__tests = tests
    self.log = gLogger.getSubLogger( 'TestExecutor' )

    if 'ResourceManagementIHEPClient' in self.apis:
      self.rmClient = self.apis[ 'ResourceManagementIHEPClient' ]
    else:
      self.rmClient = ResourceManagementIHEPClient()


  def __matchTests( self, matchArgs ):
    execTests = []

    for testType, testDict in self.__tests.items():

      testMatchArgs = testDict[ 'match' ]

      match = True
      for name, value in matchArgs.items():
        if not value:
          continue

        if type( value ) == str:
          value = ( value, )

        if not testMatchArgs.has_key( name ):
          continue

        target = testMatchArgs[ name ]
        if type( target ) == str:
          target = ( target, )

        match = False
        for val in value:
          if val in target:
            match = True
            break

        if not match:
          break

      if match:
        execTests.append( testType )

    return execTests


  def __storeTestResults( self, elementName, elementType, testResults ):
    """
      store the test results.
    """
    for testType, testDict in testResults.items():
      testDict[ 'CompletionTime' ] = testDict.get( 'CompletionTime' ) or '0000-0-0'
      testDict[ 'ApplicationTime' ] = testDict.get( 'ApplicationTime' ) or 0
      testDict[ 'JobID' ] = testDict.get( 'JobID' ) or 0


      resQuery = self.rmClient.addOrModifySAMResult(
                                                         elementName,
                                                         testType,
                                                         elementType,
                                                         testDict.get( 'Status' ),
                                                         testDict.get( 'Log' ),
                                                         testDict.get( 'JobID' ),
                                                         testDict.get( 'SubmissionTime' ),
                                                         testDict.get( 'CompletionTime' ),
                                                         testDict.get( 'ApplicationTime' ),
                                                         testDict.get( 'LastCheckTime' )
                                                         )
      if not resQuery[ 'OK' ]:
        return resQuery

    return S_OK()


  def execute( self, element, lastCheckTime = None ):
    """ Main method which executes the tests and obtains the results. Use
    two loops to do all the work. In the first loop, execute all the tests for
    corresponding elements and put the executed tests into  executedTestsQueue.
    In the second loop, traverse executedTestsQueue to obtain test results.

    examples:
      >>> executor.execute()[ 'Value' ]
          { 'Records' : ( ( 'chenj01.ihep.ac.cn', 'WMS-Test', 'ComputingElement', 'OK',
                                       'balabala', 1, '2016-5-8 00:00:00', '2016-5-8 00:05:23', 0.1234 ),
                                    ( 'chenj01.ihep.ac.cn', 'BOSS-Test', 'ComputingElement', 'Bad',
                                       'balabala', 2, '2016-5-8 00:00:00', '0000-0-0', 0 ),
                                    ( 'IHEPD-USER', 'SE-Test', 'StorageElement', 'Bad',
                                      'balabala', None, '2016-5-8 00:00:00', '0000-0-0', 0 ) ),
             'Columns' : ( 'ElementName', 'TestType', 'ElementType', 'Status',
                                      'Log', 'JobID', 'SubmissionTime', 'CompletionTime', 'ApplicationTime' ) }

    :return: S_OK( { 'Records' : `tuple`, 'Columns' : `tuple` } ) / S_ERROR
    """

    elementName = element['ElementName' ]
    elementType = element[ 'ElementType' ]
    lastCheckTime = lastCheckTime or datetime.utcnow().replace( micresecond = 0 )

    matchArgs = { 'ElementType' : elementType, 'VO' : element.get( 'VO' ) }
    execTests = self.__matchTests( matchArgs )
    if execTests == []:
      return S_ERROR( 'No SAM test matched for %s' % elementName )

    testResults = {}
    runningTestsQueue = Queue.Queue()
    for testType in execTests:
      self.log.debug( "The list of tests %s for %s" % ( testType, element[ 'ElementName' ] ) )
      testObj = self.__tests[ testType ][ 'object' ]
      result = testObj.doTest( element )
      if not result[ 'OK' ]:
        self.log.error( 'Failed to execute %s for %s' % ( testType, elementName ) )
        self.log.error( result[ 'Message' ] )
        return S_ERROR( 'Failed to execute SAM tests.' )
      result = result[ 'Value' ]
      result[ 'Result' ][ 'LastCheckTime' ] = lastCheckTime
      testResults[ testType ] = result[ 'Result' ]
      if not result[ 'Finish' ]:
        runningTestsQueue.put( testType )

    while not runningTestsQueue.empty():
      time.sleep(random.randint(30, 90))
      testType = runningTestsQueue.get_nowait()
      testObj = self.__tests[ testType ][ 'object' ]
      jobID = testResults[ testType ][ 'JobID' ]
      vo = testResults[ testType ][ 'VO' ]
      submissionTime = testResults[ testType ][ 'SubmissionTime' ]
      result = testObj.getTestResult( elementName, vo, jobID, submissionTime )
      if not result[ 'OK' ]:
        self.log.error( 'Failed to get %s result for %s' % ( testType, elementName ) )
        self.log.error( result[ 'Message' ] )
        return S_ERROR( 'Failed to get SAM test results.' )
      result = result[ 'Value' ]
      if not result:
        runningTestsQueue.put( testType )
      else:
        testResults[ testType ].update( result )
      runningTestsQueue.task_done()

    runningTestsQueue.join()

    storeRes = self.__storeTestResults( elementName, elementType, testResults )

    if not storeRes[ 'OK' ]:
      return S_ERROR( 'Failed to store SAM test results: %s' % storeRes[ 'Message' ] )

    testsStatus = {}
    for testType, testDict in testResults.items():
      testsStatus[ testType ] = testDict[ 'Status' ]
    return S_OK( testsStatus )



from DIRAC                                                         import S_OK
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.ResourceStatusSystem.Command.Command                    import Command
from IHEPDIRAC.ResourceStatusSystem.Client.ResourceManagementIHEPClient import ResourceManagementIHEPClient



class HostCommand( Command ):

  def __init__( self, args = None, clients = None ):
    super( HostCommand, self ).__init__( args, clients )

    if 'JobDB' in self.apis:
      self.jobDB = self.apis[ 'JobDB' ]
    else:
      self.jobDB = JobDB()

    if 'ResourceManagementIHEPClient' in self.apis:
      self.rmIHEPClient = self.apis[ 'ResourceManagementIHEPClient' ]
    else:
      self.rmIHEPClient = ResourceManagementIHEPClient()


  def _storeCommand( self, result):
    """
      Stores the results of doNew method on the database.
    """

    for hostDict in result:
      resQuery = self.rmClient.addOrModifyHostCache( host = hostDict[ 'Host' ],
                                                     site = hostDict[ 'Site' ],
                                                     running = hostDict[ 'Running' ],
                                                     done = hostDict[ 'Done' ],
                                                     failed = hostDict[ 'Failed' ],
                                                     efficiency = hostDict[ 'Efficiency' ] )
      if not resQuery[ 'OK' ]:
        return resQuery
    return S_OK()


  def doNew( self, masterParams = None ):

    sql = """
select JP.Value, J.Status, J.Site, count(*) from Jobs J, JobParameters JP
where J.JobID = JP.JobID and JP.Name = 'HostName'
and J.LastUpdateTime >= DATE_SUB(UTC_TIMESTAMP(),INTERVAL 24 HOUR)
group by JP.Value, J.Status
"""

    jobDB = JobDB()
    queryRes = jobDB._query(sql)
    if not queryRes[ 'OK' ]:
      return queryRes
    records = queryRes[ 'Value' ]

    hostJobs = {}

    for record in records:
      hostName = record[ 0 ]
      if hostName not in hostJobs:
        hostJobs[ hostName ] = { 'Site' : record[ 2 ], 'Running' : 0, 'Done' : 0, 'Failed' : 0 }
      hostJobs[ hostName ][ record[ 1 ] ] = record[ 3 ]

    uniformResult = []

    for host, hostDict in hostJobs.items():
      hostDict[ 'Host' ] = host
      if hostDict[ 'Done' ] == 0 and hostDict[ 'Failed' ] == 0:
        hostDict[ 'Efficiency' ] = 0
      else:
        hostDict[ 'Efficiency' ] = float( hostDict[ 'Done' ] ) / ( hostDict[ 'Done' ] + hostDict[ 'Failed' ] )
      uniformResult.append( hostDict )

    storeRes = self._storeCommand( uniformResult )
    if not storeRes[ 'OK' ]:
      return storeRes

    return S_OK( uniformResult )


  def doMaster( self ):
    """
      Master method.

      Gets all sites and calls doNew method.
    """

    jobsResults = self.doNew()
    if not jobsResults[ 'OK' ]:
      self.metrics[ 'failed' ].append( jobsResults[ 'Message' ] )

    return S_OK( self.metrics )



from DIRAC import S_OK, S_ERROR
from IHEPDIRAC.AccountingSystem.Client.Types.DataTransfer import DataTransfer
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

class DataTransferPlotter( BaseReporter ):
  _typeName = "DataTransfer"
  _typeKeyFields = [ df[0] for df in DataTransfer().definitionKeyFields]

  def _reportDataTransfered( self, reportRequest ):
    selectFields = ( self._getSelectStringForGrouping( reportRequest[ 'groupingFields' ] ) + ", SUM(%s)",
                     reportRequest[ 'groupingFields' ][1] + [ 'TransferSize' ]
                   )
    retVal = self._getSummaryData( reportRequest[ 'startTime' ],
                                reportRequest[ 'endTime' ],
                                selectFields,
                                reportRequest[ 'condDict' ],
                                reportRequest[ 'groupingFields' ],
                                {} )
    if not retVal[ 'OK' ]:
      return retVal
    dataDict = retVal[ 'Value' ]
    for key in dataDict:
      dataDict[ key ] = int( dataDict[ key ] )
    return S_OK( { 'data' : dataDict  } )

  def _plotDataTransfered( self, reportRequest, plotInfo, filename ):
    metadata = { 'title' : 'Total data transfered by %s' % reportRequest[ 'grouping' ],
                 'ylabel' : 'bytes',
                 'starttime' : reportRequest[ 'startTime' ],
                 'endtime' : reportRequest[ 'endTime' ]
                }
    return self._generatePiePlot( filename, plotInfo[ 'data'], metadata )

  def _reportThroughput( self, reportRequest ):
    selectFields = ( self._getSelectStringForGrouping( reportRequest[ 'groupingFields' ] ) + ", %s, %s, SUM(%s)",
                     reportRequest[ 'groupingFields' ][1] + [ 'startTime', 'bucketLength', 'TransferSize' ]
                   )
    retVal = self._getTimedData( reportRequest[ 'startTime' ],
                                reportRequest[ 'endTime' ],
                                selectFields,
                                reportRequest[ 'condDict' ],
                                reportRequest[ 'groupingFields' ],
                                {} )
    if not retVal[ 'OK' ]:
      return retVal
    dataDict, granularity = retVal[ 'Value' ]
    self.stripDataField( dataDict, 0 )
    dataDict, maxValue = self._divideByFactor( dataDict, granularity )
    dataDict = self._fillWithZero( granularity, reportRequest[ 'startTime' ], reportRequest[ 'endTime' ], dataDict )
    baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableRateUnit( dataDict,
                                                                                  self._getAccumulationMaxValue( dataDict ),
                                                                                  "bytes" )
    return S_OK( { 'data' : baseDataDict, 'graphDataDict' : graphDataDict,
                   'granularity' : granularity, 'unit' : unitName } )

  def _plotThroughput( self, reportRequest, plotInfo, filename ):
    metadata = { 'title' : 'Throughput by %s' % reportRequest[ 'grouping' ] ,
                 'ylabel' : plotInfo[ 'unit' ],
                 'starttime' : reportRequest[ 'startTime' ],
                 'endtime' : reportRequest[ 'endTime' ],
                 'span' : plotInfo[ 'granularity' ] }
    return self._generateTimedStackedBarPlot( filename, plotInfo[ 'graphDataDict' ], metadata )

  # Generate Transfer Data Accumulate
  def _reportDataSizeBin( self, reportRequest ):
    selectFields = ( self._getSelectStringForGrouping( reportRequest[ 'groupingFields' ] ) + ", %s, %s, SUM(%s)",
                     reportRequest[ 'groupingFields' ][1] + [ 'startTime', 'bucketLength',
                                    'TransferSize'
                                   ]
                   )
    retVal = self._getTimedData( reportRequest[ 'startTime' ],
                                reportRequest[ 'endTime' ],
                                selectFields,
                                reportRequest[ 'condDict' ],
                                reportRequest[ 'groupingFields' ],
                                {} )
    if not retVal[ 'OK' ]:
      return retVal
    dataDict, granularity = retVal[ 'Value' ]
    self.stripDataField( dataDict, 0 )
    dataDict = self._fillWithZero( granularity, reportRequest[ 'startTime' ], reportRequest[ 'endTime' ], dataDict )
    #dataDict = self._accumulate( granularity, reportRequest[ 'startTime' ], reportRequest[ 'endTime' ], dataDict )
    baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableUnit( dataDict,
                                                                              self._getAccumulationMaxValue( dataDict ),
                                                                              "bytes" )
    return S_OK( { 'data' : baseDataDict, 'graphDataDict' : graphDataDict,
                   'granularity' : granularity, 'unit' : unitName } )

  def _plotDataSizeBin( self, reportRequest, plotInfo, filename ):
    metadata = { 'title' : 'Transfered data by %s' % reportRequest[ 'grouping' ] ,
                 'starttime' : reportRequest[ 'startTime' ],
                 'endtime' : reportRequest[ 'endTime' ],
                 'span' : plotInfo[ 'granularity' ],
                 'ylabel' : plotInfo[ 'unit' ],
                 'sort_labels' : 'last_value' }
    return self._generateTimedStackedBarPlot( filename, plotInfo[ 'graphDataDict' ], metadata )

  # Quality
  def _reportQuality( self, reportRequest ):
    selectFields = ( self._getSelectStringForGrouping( reportRequest[ 'groupingFields' ] ) + ", %s, %s, SUM(%s), SUM(%s)",
                     reportRequest[ 'groupingFields' ][1] + [ 'startTime', 'bucketLength',
                                    'TransferOK', 'TransferTotal'
                                   ]
                   )
    retVal = self._getTimedData( reportRequest[ 'startTime' ],
                                reportRequest[ 'endTime' ],
                                selectFields,
                                reportRequest[ 'condDict' ],
                                reportRequest[ 'groupingFields' ],
                                { 'checkNone' : True,
                                  'convertToGranularity' : 'sum',
                                  'calculateProportionalGauges' : False,
                                  'consolidationFunction' : self._efficiencyConsolidation } )
    if not retVal[ 'OK' ]:
      return retVal
    dataDict, granularity = retVal[ 'Value' ]
    self.stripDataField( dataDict, 0 )
    if len( dataDict ) > 1:
      #Get the total for the plot
      selectFields = ( "'Total', %s, %s, SUM(%s),SUM(%s)",
                       [ 'startTime', 'bucketLength',
                         'TransferOK', 'TransferTotal'
                       ]
                     )
      retVal = self._getTimedData( reportRequest[ 'startTime' ],
                                  reportRequest[ 'endTime' ],
                                  selectFields,
                                  reportRequest[ 'condDict' ],
                                  reportRequest[ 'groupingFields' ],
                                  { 'checkNone' : True,
                                    'convertToGranularity' : 'sum',
                                    'calculateProportionalGauges' : False,
                                    'consolidationFunction' : self._efficiencyConsolidation } )
      if not retVal[ 'OK' ]:
        return retVal
      totalDict = retVal[ 'Value' ][0]
      self.stripDataField( totalDict, 0 )
      for key in totalDict:
        dataDict[ key ] = totalDict[ key ]
    return S_OK( { 'data' : dataDict, 'granularity' : granularity } )

  def _plotQuality( self, reportRequest, plotInfo, filename ):
    metadata = { 'title' : 'Transfer quality by %s' % reportRequest[ 'grouping' ] ,
                 'starttime' : reportRequest[ 'startTime' ],
                 'endtime' : reportRequest[ 'endTime' ],
                 'span' : plotInfo[ 'granularity' ] }
    return self._generateQualityPlot( filename, plotInfo[ 'data' ], metadata )



""" JUNOTest

  A test class to test the software for the vo cepc.

"""

from IHEPDIRAC.ResourceStatusSystem.SAM.SAMTest.CEBaseTest import CEBaseTest


__RCSID__ = '$Id: $'


class CEPCTest( CEBaseTest ):
  """
    JUNOTest is used to test whether the cepc's software is fine to run jobs.
  """

  def __init__( self, args = None, apis = None ):
    super( CEPCTest, self ).__init__( args, apis )


  @staticmethod
  def _judge( log ):
    idx = log.find( 'Job Done.' )
    if idx != -1:
      return 'OK'
    else:
      return 'Bad'



''' ResourceManagementHandler

  Module that allows users to access the ResourceManagementDB remotely.

'''

from DIRAC                                              import gConfig, S_OK, gLogger
from DIRAC.Core.DISET.RequestHandler                    import RequestHandler
from DIRAC.ResourceStatusSystem.Utilities               import Synchronizer, Utils
from IHEPDIRAC.ResourceStatusSystem.DB.ResourceManagementIHEPDB import ResourceManagementIHEPDB

__RCSID__ = '$Id: $'

def initializeResourceManagementIHEPHandler( _serviceInfo ):
  '''
    Handler initialization, where we set the ResourceManagementDB as global db.
  '''

  global db
  db = ResourceManagementIHEPDB()
  # Regenerates DB tables if needed
  db._checkTable()

  syncObject = Synchronizer.Synchronizer()
  gConfig.addListenerToNewVersionEvent( syncObject.sync )

  return S_OK()

################################################################################

class ResourceManagementIHEPHandler( RequestHandler ):
  '''
  The ResourceManagementHandler exposes the DB front-end functions through a
  XML-RPC server, functionalities inherited from :class:`DIRAC.Core.DISET.Reques\
  tHandler.RequestHandler`

  According to the ResourceManagementDB philosophy, only functions of the type:
  - insert
  - update
  - select
  - delete

  are exposed. If you need anything more complicated, either look for it on the
  :class:`ResourceManagementClient`, or code it yourself. This way the DB and the
  Service are kept clean and tidied.

  To can use this service on this way, but you MUST NOT DO IT. Use it through the
  :class:`ResourceManagementClient`. If offers in the worst case as good perfor\
  mance as the :class:`ResourceManagementHandler`, if not better.

   >>> from DIRAC.Core.DISET.RPCClient import RPCCLient
   >>> server = RPCCLient("ResourceStatus/ResourceManagement")
  '''

  def __init__( self, *args, **kwargs ):
    super( ResourceManagementIHEPHandler, self ).__init__( *args, **kwargs )

  @staticmethod
  def __logResult( methodName, result ):
    '''
      Method that writes to log error messages
    '''

    if not result[ 'OK' ]:
      gLogger.error( '%s%s' % ( methodName, result[ 'Message' ] ) )

  @staticmethod
  def setDatabase( database ):
    '''
    This method let us inherit from this class and overwrite the database object
    without having problems with the global variables.

    :Parameters:
      **database** - `MySQL`
        database used by this handler

    :return: None
    '''
    global db
    db = database

  types_insert = [ dict, dict ]
  def export_insert( self, params, meta ):
    '''
    This method is a bridge to access :class:`ResourceManagementDB` remotely. It
    does not add neither processing nor validation. If you need to know more
    about this method, you must keep reading on the database documentation.

    :Parameters:
      **params** - `dict`
        arguments for the mysql query ( must match table columns ! ).

      **meta** - `dict`
        metadata for the mysql query. It must contain, at least, `table` key
        with the proper table name.

    :return: S_OK() || S_ERROR()
    '''

    gLogger.info( 'insert: %s %s' % ( params, meta ) )

    res = db.insert( params, meta )
    self.__logResult( 'insert', res )

    return res

  types_update = [ dict, dict ]
  def export_update( self, params, meta ):
    '''
    This method is a bridge to access :class:`ResourceManagementDB` remotely. It
    does not add neither processing nor validation. If you need to know more
    about this method, you must keep reading on the database documentation.

    :Parameters:
      **params** - `dict`
        arguments for the mysql query ( must match table columns ! ).

      **meta** - `dict`
        metadata for the mysql query. It must contain, at least, `table` key
        with the proper table name.

    :return: S_OK() || S_ERROR()
    '''

    gLogger.info( 'update: %s %s' % ( params, meta ) )

    res = db.update( params, meta )
    self.__logResult( 'update', res )

    return res

  types_select = [ dict, dict ]
  def export_select( self, params, meta ):
    '''
    This method is a bridge to access :class:`ResourceManagementDB` remotely.
    It does not add neither processing nor validation. If you need to know more\
    about this method, you must keep reading on the database documentation.

    :Parameters:
      **params** - `dict`
        arguments for the mysql query ( must match table columns ! ).

      **meta** - `dict`
        metadata for the mysql query. It must contain, at least, `table` key
        with the proper table name.

    :return: S_OK() || S_ERROR()
    '''

    gLogger.info( 'select: %s %s' % ( params, meta ) )


    res = db.select( params, meta )
    self.__logResult( 'select', res )

    return res

  types_delete = [ dict, dict ]
  def export_delete( self, params, meta ):
    '''
    This method is a bridge to access :class:`ResourceManagementDB` remotely.\
    It does not add neither processing nor validation. If you need to know more \
    about this method, you must keep reading on the database documentation.

    :Parameters:
      **params** - `dict`
        arguments for the mysql query ( must match table columns ! ).

      **meta** - `dict`
        metadata for the mysql query. It must contain, at least, `table` key
        with the proper table name.

    :return: S_OK() || S_ERROR()
    '''

    gLogger.info( 'delete: %s %s' % ( params, meta ) )

    res = db.delete( params, meta )
    self.__logResult( 'delete', res )

    return res

  types_addOrModify = [ dict, dict ]
  def export_addOrModify( self, params, meta ):
    '''
    This method is a bridge to access :class:`ResourceManagementDB` remotely. It does
    not add neither processing nor validation. If you need to know more about
    this method, you must keep reading on the database documentation.

    :Parameters:
      **args** - `tuple`
        arguments for the mysql query ( must match table columns ! ).

      **kwargs** - `dict`
        metadata for the mysql query. It must contain, at least, `table` key
        with the proper table name.

    :return: S_OK() || S_ERROR()
    '''

    gLogger.info( 'addOrModify: %s %s' % ( params, meta ) )

    res = db.addOrModify( params, meta )
    self.__logResult( 'addOrModify', res )

    return res

  types_addIfNotThere = [ dict, dict ]
  def export_addIfNotThere( self, params, meta ):
    '''
    This method is a bridge to access :class:`ResourceManagementDB` remotely. It does
    not add neither processing nor validation. If you need to know more about
    this method, you must keep reading on the database documentation.

    :Parameters:
      **args** - `tuple`
        arguments for the mysql query ( must match table columns ! ).

      **kwargs** - `dict`
        metadata for the mysql query. It must contain, at least, `table` key
        with the proper table name.

    :return: S_OK() || S_ERROR()
    '''

    gLogger.info( 'addIfNotThere: %s %s' % ( params, meta ) )

    res = db.addIfNotThere( params, meta )
    self.__logResult( 'addIfNotThere', res )

    return res

################################################################################
#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



#!/usr/bin/env python
from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType

class DataTransfer(BaseAccountingType):

  def __init__(self):
    BaseAccountingType.__init__(self)

    self.definitionKeyFields = [
                                ( 'User', 'VARCHAR(32)'),
                                ( 'Source', 'VARCHAR(32)'),
                                ( 'Destination', 'VARCHAR(32)'),
                                ( 'Protocol', 'VARCHAR(32)'),
                                ( 'FinalStatus', 'VARCHAR(32)')
                               ]
    self.definitionAccountingFields = [ 
                                        ( 'TransferSize', 'BIGINT UNSIGNED' ),
                                        ( 'TransferTime', 'FLOAT' ),
                                        ( 'TransferOK', 'INT UNSIGNED' ),
                                        ( 'TransferTotal', 'INT UNSIGNED' )
                                      ]

    self.checkType()



""" SETest

  A test class to test the availability of SE.

"""

import os, time
from datetime                                           import datetime
from DIRAC                                              import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Subprocess                    import systemCall
from DIRAC.DataManagementSystem.Client.DataManager      import DataManager
from IHEPDIRAC.ResourceStatusSystem.SAM.SAMTest.TestBase import TestBase
from IHEPDIRAC.ResourceStatusSystem.Utilities            import BESUtils



class SETest( TestBase ):
  """
    SETest is used to test the availability of SE.
  """

  def __init__( self, args = None, apis = None ):
    super( SETest, self ).__init__( args, apis )

    self.timeout = self.args.get( 'timeout', 60 )
    self.__lfnPath = '/{vo}/user/z/zhangxm/'
    self.__testFile = 'test.dat'
    self.__localPath = '/tmp/'
    self.__scriptPath = '/opt/dirac/pro/IHEPDIRAC/ResourceStatusSystem/SAM/sam_script'
    self.__scriptName = 'se_test.py'

    if 'DataManager' in self.apis:
      self.dm = self.apis[ 'DataManager' ]
    else:
      self.dm = DataManager()


  def doTest( self, elementDict ):
    """
      Test upload and download for specified SE.
    """

    elementName = elementDict[ 'ElementName' ]
    vo = elementDict[ 'VO' ]

    testFilePath = self.__localPath + self.__testFile
    if not os.path.exists( testFilePath ) or not os.path.isfile( testFilePath ):
      f = open( testFilePath, 'w' )
      f.write( 'hello' )
      f.close()

    status = 'OK'
    log = ''
    lfnPath = self.__lfnPath.format(vo=vo) + elementName + '-' + self.__testFile
    submissionTime = datetime.utcnow().replace( microsecond = 0 )

    proxyPath = BESUtils.getProxyByVO( 'zhangxm', vo )
    if not proxyPath[ 'OK' ]:
      gLogger.error('Can not get proxy for VO %s' % vo)
      return proxyPath
    proxyPath = proxyPath[ 'Value' ]

    env_test = os.environ.copy()
    env_test[ 'X509_USER_PROXY' ] = proxyPath
    cmd = [os.path.join(self.__scriptPath, self.__scriptName), '-o', '/DIRAC/Security/UseServerCertificate=no', lfnPath, testFilePath, elementName]
    result = systemCall(300, cmd, env=env_test)
    print result
    if not result['OK']:
      status = 'Bad'
      log += 'Call %s failed: %s' % (self.__scriptName, result['Message'])
    elif result['Value'][0] != 0:
      status = 'Bad'
      log += '%s exit with error %s:\n%s' % (self.__scriptName, result['Value'][0], result['Value'][1])
    else:
      log += '%s exit successfully:\n%s' % (self.__scriptName, result['Value'][1])

    completionTime = datetime.utcnow().replace( microsecond = 0 )
    applicationTime = ( completionTime - submissionTime ).total_seconds()

    result = { 'Result' : { 'Status' : status,
                            'Log' : log,
                            'SubmissionTime' : submissionTime,
                            'CompletionTime' : completionTime,
                            'ApplicationTime' : applicationTime },
               'Finish' : True }

#    if os.path.exists( testFilePath ) and os.path.isfile( testFilePath ):
#      os.remove( testFilePath )
    localFile = self.__localPath + elementName +'-' + self.__testFile
    if os.path.exists( localFile ) and os.path.isfile( localFile ):
      os.remove( localFile )

    return S_OK( result )




""" CSHelpers

  Module containing functions interacting with the CS and useful for the RSS
  modules.
"""

__RCSID__ = '2a2cb06 (2019-08-21 01:31:04 +0800) Xianghu Zhao <zhaoxh@ihep.ac.cn>'

from DIRAC import gConfig, gLogger, S_OK
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.Core.Utilities.SiteSEMapping import getSEParameters
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getQueues
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.ResourceStatusSystem.Utilities import Utils


def warmUp():
  """
    gConfig has its own dark side, it needs some warm up phase.
  """
  from DIRAC.ConfigurationSystem.private.Refresher import gRefresher
  gRefresher.refreshConfigurationIfNeeded()


#@deprecated("Use DIRAC.ConfigurationSystem.Client.Helpers.Resources.getSites() instead")
def getSites():
  """
    Gets all sites from /Resources/Sites
  """

  _basePath = 'Resources/Sites'

  sites = []

  domainNames = gConfig.getSections(_basePath)
  if not domainNames['OK']:
    return domainNames
  domainNames = domainNames['Value']

  for domainName in domainNames:
    domainSites = gConfig.getSections('%s/%s' % (_basePath, domainName))
    if not domainSites['OK']:
      return domainSites

    domainSites = domainSites['Value']

    sites.extend(domainSites)

  # Remove duplicated ( just in case )
  sites = list(set(sites))
  return S_OK(sites)


@deprecated("unused")
def getDomainSites():
  """
    Gets all sites from /Resources/Sites
  """

  _basePath = 'Resources/Sites'

  sites = {}

  domainNames = gConfig.getSections(_basePath)
  if not domainNames['OK']:
    return domainNames
  domainNames = domainNames['Value']

  for domainName in domainNames:
    domainSites = gConfig.getSections('%s/%s' % (_basePath, domainName))
    if not domainSites['OK']:
      return domainSites

    domainSites = domainSites['Value']

    sites[domainName] = domainSites

  return S_OK(sites)


def getResources():
  """
    Gets all resources
  """

  resources = DMSHelpers().getStorageElements()

  fts = getFTS()
  if fts['OK']:
    resources = resources + fts['Value']

  fc = getFileCatalogs()
  if fc['OK']:
    resources = resources + fc['Value']

  ce = getComputingElements()
  if ce['OK']:
    resources = resources + ce['Value']

  return S_OK(resources)


@deprecated("unused")
def getNodes():
  """
    Gets all nodes
  """

  nodes = []

  queues = getQueues()
  if queues['OK']:
    nodes = nodes + queues['Value']

  return S_OK(nodes)


#@deprecated("unused")
def getStorageElements():
  """
    Gets all storage elements from /Resources/StorageElements
  """

  _basePath = 'Resources/StorageElements'

  seNames = gConfig.getSections(_basePath)
  return seNames


@deprecated("unused")
def getSEToken(seName):
  """ Get StorageElement token
  """

  seParameters = getSEParameters(seName)
  if not seParameters['OK']:
    gLogger.warn("Could not get SE parameters", "SE: %s" % seName)
    return seParameters

  return S_OK(seParameters['Value']['SpaceToken'])


def getStorageElementEndpoint(seName):
  """ Get endpoints of a StorageElement

      :param str seName: name of the storage element

      :returns: S_OK() or S_ERROR
                for historical reasons, if the protocol is SRM, you get  'httpg://host:port/WSUrl'
                For other protocols, you get :py:meth:`~DIRAC.Resources.Storage.StorageBase.StorageBase.getEndpoint`

  """
  seParameters = getSEParameters(seName)
  if not seParameters['OK']:
    gLogger.warn("Could not get SE parameters", "for SE %s" % seName)
    return seParameters

  seEndpoints = []

  for parameters in seParameters['Value']:
    if parameters['Protocol'].lower() == 'srm':
      # we need to construct the URL with httpg://
      host = parameters['Host']
      port = parameters['Port']
      wsurl = parameters['WSUrl']
      # MAYBE wusrl is not defined
      if host and port:
        url = 'httpg://%s:%s%s' % (host, port, wsurl)
        url = url.replace('?SFN=', '')
        seEndpoints.append(url)
    else:
      seEndpoints.append(parameters['Endpoint'])

  return S_OK(seEndpoints)


@deprecated("unused")
def getStorageElementEndpoints(storageElements=None):
  """ get the endpoints of the Storage ELements
  """

  if storageElements is None:
    storageElements = DMSHelpers().getStorageElements()

  storageElementEndpoints = []

  for se in storageElements:

    seEndpoint = getStorageElementEndpoint(se)
    if not seEndpoint['OK']:
      continue
    storageElementEndpoints.append(seEndpoint['Value'])

  return S_OK(list(set(storageElementEndpoints)))


def getFTS():
  """
    Gets all FTS endpoints
  """

  # FIXME: FTS2 will be deprecated (first 2 lines that follow)
  ftsEndpoints = gConfig.getValue('Resources/FTSEndpoints/Default/FTSEndpoint', [])
  ftsEndpoints += _getFTSEndpoints('Resources/FTSEndpoints/FTS2')
  ftsEndpoints += _getFTSEndpoints()

  return S_OK(ftsEndpoints)


def _getFTSEndpoints(basePath='Resources/FTSEndpoints/FTS3'):
  """
    Gets all FTS endpoints that are in CS
  """

  result = gConfig.getOptions(basePath)
  if result['OK']:
    return result['Value']
  return []


def getSpaceTokenEndpoints():
  """ Get Space Token Endpoints """

  return Utils.getCSTree('Shares/Disk')


def getFileCatalogs():
  """
    Gets all storage elements from /Resources/FileCatalogs
  """

  _basePath = 'Resources/FileCatalogs'

  fileCatalogs = gConfig.getSections(_basePath)
  return fileCatalogs


def getComputingElements():
  """
    Gets all computing elements from /Resources/Sites/<>/<>/CE
  """
  _basePath = 'Resources/Sites'

  ces = []

  domainNames = gConfig.getSections(_basePath)
  if not domainNames['OK']:
    return domainNames
  domainNames = domainNames['Value']

  for domainName in domainNames:
    domainSites = gConfig.getSections('%s/%s' % (_basePath, domainName))
    if not domainSites['OK']:
      return domainSites
    domainSites = domainSites['Value']

    for site in domainSites:
      siteCEs = gConfig.getSections('%s/%s/%s/CEs' % (_basePath, domainName, site))
      if not siteCEs['OK']:
        # return siteCEs
        gLogger.error(siteCEs['Message'])
        continue
      siteCEs = siteCEs['Value']
      ces.extend(siteCEs)

  # Remove duplicated ( just in case )
  ces = list(set(ces))

  return S_OK(ces)


def getSiteComputingElements(siteName):
  """
    Gets all computing elements from /Resources/Sites/<>/<siteName>/CE
  """

  _basePath = 'Resources/Sites'

  domainNames = gConfig.getSections(_basePath)
  if not domainNames['OK']:
    return domainNames
  domainNames = domainNames['Value']

  for domainName in domainNames:
    ces = gConfig.getValue('%s/%s/%s/CE' % (_basePath, domainName, siteName), '')
    if ces:
      return ces.split(', ')

  return []


#@deprecated("Use DIRAC.Core.Utilities.SiteSEMapping.getSEsForSite() instead")
def getSiteStorageElements(siteName):
  """
    Gets all storage elements from /Resources/Sites/<>/<siteName>/SE

      Used by WebApp/SiteSummaryHandler.py
  """

  _basePath = 'Resources/Sites'

  domainNames = gConfig.getSections(_basePath)
  if not domainNames['OK']:
    return domainNames
  domainNames = domainNames['Value']

  for domainName in domainNames:
    ses = gConfig.getValue('%s/%s/%s/SE' % (_basePath, domainName, siteName), '')
    if ses:
      return ses.split(', ')

  return []


def getSiteElements(siteName):
  """
    Gets all the computing and storage elements for a given site
  """

  res = DMSHelpers().getSiteSEMapping()
  if not res['OK']:
    return res
  resources = res['Value'][1].get(siteName, [])

  res = getQueues(siteName)
  if not res['OK']:
    return res
  resources = list(resources) + list(res['Value'][siteName])

  return S_OK(resources)


def getQueuesRSS():
  """
    Gets all computing elements from /Resources/Sites/<>/<>/CE/Queues
  """
  _basePath = 'Resources/Sites'

  queues = []

  domainNames = gConfig.getSections(_basePath)
  if not domainNames['OK']:
    return domainNames
  domainNames = domainNames['Value']

  for domainName in domainNames:
    domainSites = gConfig.getSections('%s/%s' % (_basePath, domainName))
    if not domainSites['OK']:
      return domainSites
    domainSites = domainSites['Value']

    for site in domainSites:
      siteCEs = gConfig.getSections('%s/%s/%s/CEs' % (_basePath, domainName, site))
      if not siteCEs['OK']:
        # return siteCEs
        gLogger.error(siteCEs['Message'])
        continue
      siteCEs = siteCEs['Value']

      for siteCE in siteCEs:
        siteQueue = gConfig.getSections('%s/%s/%s/CEs/%s/Queues' % (_basePath, domainName, site, siteCE))
        if not siteQueue['OK']:
          # return siteQueue
          gLogger.error(siteQueue['Message'])
          continue
        siteQueue = siteQueue['Value']

        queues.extend(siteQueue)

  # Remove duplicated ( just in case )
  queues = list(set(queues))

  return S_OK(queues)


@deprecated("unused")
def getRegistryUsers():
  """
    Gets all users from /Registry/Users
  """

  _basePath = 'Registry/Users'

  registryUsers = {}

  userNames = gConfig.getSections(_basePath)
  if not userNames['OK']:
    return userNames
  userNames = userNames['Value']

  for userName in userNames:

    # returns { 'Email' : x, 'DN': y, 'CA' : z }
    userDetails = gConfig.getOptionsDict('%s/%s' % (_basePath, userName))
    if not userDetails['OK']:
      return userDetails

    registryUsers[userName] = userDetails['Value']

  return S_OK(registryUsers)



# -*- coding: utf-8 -*-

import datetime
import random

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from IHEPDIRAC.TransferSystem.DB.TransferDB import TransRequestEntryWithID
from IHEPDIRAC.TransferSystem.DB.TransferDB import TransFileListEntryWithID

from IHEPDIRAC.TransferSystem.Agent.helper.TransferFactory import gTransferFactory

from IHEPDIRAC.AccountingSystem.Client.Types.DataTransfer import DataTransfer

class helper_TransferAgent(object):

  def __init__(self, transferAgent, gTransferDB):
    self.transferAgent =transferAgent
    self.transferDB = gTransferDB
    gLogger.info("Creating File Catalog")
    self.fileCatalog = FileCatalog()

  def helper_add_transfer(self, result):
    if not result:
      gLogger.error("There is no infomation")
      return False

    res = self.transferDB.get_TransferRequest(condDict={
                                                      "id": result.trans_req_id
                                                      })
    if not res["OK"]:
      return False
    req_list = res["Value"]
    if len(req_list) != 1:
      return False
    req =  TransRequestEntryWithID._make(req_list[0])

    # construct the info
    info = {"id": result.id,
            "LFN": result.LFN,
            "srcSE": req.srcSE,
            "dstSE": req.dstSE,
            "retransfer": -1,
            "error": ""}
    # Add the Transfer
    worker = gTransferFactory.generate(req.protocol, info)
    if worker is None:
      return True
    self.transferAgent.transfer_worker.append(worker)
    # Change the status
    self.helper_status_update(
        self.transferDB.tables["TransferFileList"],
        result.id,
        {"status":"transfer", 
          "start_time":datetime.datetime.utcnow()})
    # Add Accounting:
    d = {}
    d["User"] = req.username
    d["Source"] = req.srcSE
    d["Destination"] = req.dstSE
    d["Protocol"] = req.protocol
    d["FinalStatus"] = "OK"
    d["TransferSize"] = 0 # TODO
    r = self.fileCatalog.getFileSize(result.LFN)
    if r["OK"]:
      if r["Value"]["Successful"]:
        d["TransferSize"] = r["Value"]["Successful"][result.LFN]
    d["TransferTime"] = 1 # 1s 
    d["TransferOK"] = 1
    d["TransferTotal"] = 1
    acct_dt = DataTransfer()
    acct_dt.setValuesFromDict(d)
    acct_dt.setNowAsStartAndEndTime()
    # save it 
    worker.acct_dt = acct_dt

    return True

  def helper_remove_transfer(self, worker):
    info = worker.info
    gLogger.info("File.id = %d -> finish" % info["id"])
    self.helper_status_update(
        self.transferDB.tables["TransferFileList"],
        info["id"],
        {"status":"finish", 
          "finish_time": datetime.datetime.utcnow()})
    # Accounting
    acct_dt = worker.acct_dt
    acct_dt.setEndTime()
    # TODO
    d = {}
    td = acct_dt.endTime-acct_dt.startTime
    td_s = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
    d["TransferTime"] = td_s # 1s 
    if info["error"]:
      d["FinalStatus"] = "FAILED"
      d["TransferOK"] = 0
    else:  
      d["FinalStatus"] = "OK"
      d["TransferOK"] = 1

    acct_dt.setValuesFromDict(d)

    acct_dt.commit()
    gLogger.info("Submit Accounting Data")
    
  def helper_check_request(self):
    """
      check if the *transfer* request are ok.
      if the whole files are *finish*, then this request
      will become *finish*.
    """
    infoDict = {"status": "transfer"}
    res = self.transferDB.get_TransferRequest(condDict = infoDict)
    if not res["OK"]:
      return
    reqlist = map(TransRequestEntryWithID._make, res["Value"])
    for req in reqlist:
      res = self.transferDB._query(
          'select count(*) from %(table)s where trans_req_id = %(id)d and status not in %(status_list)s' % {
             "table": self.transferDB.tables["TransferFileList"], 
             "id": req.id,
             "status_list": '("finish", "kill")' # XXX finish or kill means this request is ok.
             }
          )
      if not res["OK"]:
        # TODO
        continue
      count = res["Value"][0][0]
      if count == 0:
        # if all status is finish,
        # the req status --> finish
        gLogger.info("req.id %d change from %s to finish" % (req.id, req.status))
        self.helper_status_update(
            self.transferDB.tables["TransferRequest"],
            req.id,
            {"status":"finish"})
    return 

  def helper_get_new_request(self):
    # 1. get the *new* File in the <<Transfer File List>>.
    #    if we get, goto <<Add New Transfer>>
    already_load_status = False
    result_new_file = self.helper_get_new_File()
    # 1.1 2014.04.20
    #     They want to the other requests are also loaded,
    #     so I have to not return immediately
    if result_new_file:
      already_load_status = True
    # 2. if we can't get, use should get a *new* request
    #    from the <<Transfer Request>>.
    #    if we can't get, return False. STOP
    self.helper_check_request()
    result = self.helper_get_new_request_entry()
    if result:
      # 3. add the filelist in the dataset to the << Transfer File List >>
      condDict = {"name":result.dataset}  
      res = self.transferDB.get_Dataset(condDict)
      if not res["OK"]:
        gLogger.error(res)
        return None
      filelist = res["Value"]
      # update the status in << Request >>
      if len(filelist) > 0:
        req_status = "transfer"
      else:
        req_status = "finish"
      self.helper_status_update(self.transferDB.tables["TransferRequest"],
                                result.id,
                                {"status":req_status})
      self.transferDB.insert_TransferFileList(result.id, filelist)
    # 4. get the *new* File Again.
    # 5. can't get, return False. STOP
    # 4.prelude
    #    If already loaded, return the last result
    if already_load_status and result_new_file:
      return result_new_file
      
    result = self.helper_get_new_File()
    return result

  def helper_get_new_request_entry(self):
    """
    TransRequestEntryWithID(
      id=1L, 
      username='lintao', 
      dataset='my-dataset', 
      srcSE='IHEP-USER', 
      dstSE='IHEPD-USER', 
      submit_time=datetime.datetime(2013, 3, 13, 20, 9, 34), 
      status='new')
    """
    condDict = {"status": "new"}
    res = self.transferDB.get_TransferRequest(condDict)
    if not res["OK"]:
      return None
    req_list = res["Value"]
    len_req = len(req_list)
    if len_req:
      # random select
      tmp_idx = random.randint(0, len_req-1)
      return TransRequestEntryWithID._make(req_list[tmp_idx])
    pass

  def helper_get_new_File(self):
    """
    >>> helper.helper_get_new_File()
    TransFileListEntryWithID(
      id=1L, 
      LFN='/path/does/not/exist', 
      trans_req_id=1L, 
      start_time=None, 
      finish_time=None, 
      status='new')
    """
    condDict = {"status": "new"}
    res = self.transferDB.get_TransferFileList(condDict)
    if not res["OK"]:
      gLogger.error(res)
      return None
    filelist = res["Value"]
    gLogger.info("Filelist:")
    gLogger.info(filelist)
    len_files = len(filelist)
    if  len_files > 0:
      tmp_idx = random.randint(0, len_files-1)
      gLogger.info("get file entry index randomly: %d/%d"%(tmp_idx, len_files))
      gLogger.info("get file entry", filelist[tmp_idx])
      return TransFileListEntryWithID._make(filelist[tmp_idx])
    return None

  def helper_status_update(self, table, id, toUpdate):
    res = self.transferDB.updateFields(
                              table,
                              updateDict = toUpdate,
                              condDict = {"id":id},
                              )
    print res

  def helper_error_report(self, worker, reason):
    self.helper_status_update(self.transferDB.tables["TransferFileList"],
                              worker.info["id"],
                              {"error": reason})

  def check_worker_status(self, worker):
    """check whether the file transfer is kill(in DB)"""
    res = self.transferDB.getFields(self.transferDB.tables["TransferFileList"],
                                    outFields = ["status"],
                                    condDict = {"id":worker.info["id"]})
    if not res["OK"]:
      gLogger.error(res)
      return

    if not res["Value"]:
      return

    if len(res["Value"]) != 1:
      gLogger.error[res]
      return 

    status = res["Value"][0][0]
    if status == "kill":
      gLogger.info("check worker should be killed: ", status)
      worker.proc.kill()

if __name__ == "__main__":
  from DIRAC.Core.Base import Script
  Script.parseCommandLine( ignoreErrors = True )

  from IHEPDIRAC.TransferSystem.DB.TransferDB import TransferDB 
  gTransferDB = TransferDB()
  transferAgent = gTransferDB
  transferAgent.transfer_worker = []
  helper = helper_TransferAgent(transferAgent, gTransferDB)
  entry = helper.helper_get_new_File()
  print helper.helper_get_new_request_entry()

  if entry:
    print helper.helper_status_update( table = "TransferFileList", 
                                       id = entry.id,
                                       toStatus = {"status":"transfer"})

  print helper.helper_check_request()
  print helper.helper_add_transfer(entry)
  print transferAgent.transfer_worker

  pass



import os
import sys
import time
import datetime

from DIRAC import gLogger

from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

class GetFile(object):
  def __init__(self):
    super(GetFile, self).__init__()

    # Cache for remote attributes
    self._remoteAttributes = {}

    # These attributes could be reset in the derived classes
    self._directlyRead = False
    self._localValidation = True
    self._checksumType = 'Md5'

################################################################################
# These methods should be implemented in the derived classes

  def _available(self):
    raise Exception('not implemented')

  def _downloadSingleFile(self, remotePath, localPath):
    raise Exception('not implemented')


# These methods could be implemented optionally

  def _lfnToRemote(self, lfn):
    return lfn


################################################################################
# Common methods for GetFile

  def available(self):
    return self._available()

  def lfnToRemote(self, lfn):
    return self._lfnToRemote(lfn)

  def lfnToLocal(self, dir, lfn):
    return self.__lfnToLocal(dir, lfn)

  def setUseChecksum(self, useChecksum = True):
    self._useChecksum = useChecksum

  def setLocalValidation(self, localValidation = True):
    self._localValidation = localValidation

  def directlyRead(self):
    return self._directlyRead

  def getAllRemoteAttributes(self, lfnList):
    return self.__getAllRemoteAttribute(lfnList)

  def getRemoteAttribute(self, lfn):
    return self.__getRemoteAttribute(lfn)

  def getFile(self, lfn, dir):
    return self.__getFile(lfn, dir)


################################################################################
# Private methods for GetFile

  def __lfnToLocal(self, dir, lfn):
    return os.path.join(dir, os.path.basename(lfn))

  def __getFile(self, lfn, dir):
    result = {}

    remotePath = self._lfnToRemote(lfn)
    localPath = os.path.join(dir, os.path.basename(lfn))

    remoteAttribute = self.__getRemoteAttribute(lfn)
    if not remoteAttribute:
      gLogger.debug('Remote file does not exist:', remotePath)
      result['status'] = 'notexist'
      return result

    result['size'] = remoteAttribute['size']

    if self._localValidation:
      if self.__localValid(lfn, localPath):
        gLogger.debug('Skip downloading %s. %s already exists' % (remotePath, localPath))
        result['status'] = 'skip'
        return result

    self.__removeLocal(localPath)

    startTime = time.time()
    ret = self._downloadSingleFile(remotePath, localPath)
    endTime = time.time()
    result['span'] = endTime - startTime

    if not ret:
      gLogger.debug('Download error: %s -> %s' % (remotePath, localPath))
      result['status'] = 'error'
      return result

    self.__setFileTime(localPath, remoteAttribute['time'])
    result['status'] = 'ok'

    return result


  # not used
  def __retrieveRemoteAttribute(self, remotePath):
    ''' Get size, time and optional checksum
    '''
    raise Exception('not implemented')

  def __retrieveAllRemoteAttributes(self, lfnList):
    fc = FileCatalogClient('DataManagement/FileCatalog')
    result = fc.getFileMetadata(lfnList)
    if not result['OK']:
      raise Exception('getFileMetadata failed: %s' % result['Message'])

    attributes = {}
    for lfn in lfnList:
      if lfn in result['Value']['Successful']:
        attributes[lfn] = self.__parseMetadata(result['Value']['Successful'][lfn])

    return attributes


  def __getAllRemoteAttribute(self, lfnList):
    self._remoteAttributes = self.__retrieveAllRemoteAttributes(lfnList)
    return self._remoteAttributes

  def __getRemoteAttribute(self, lfn):
    if lfn in self._remoteAttributes:
      return self._remoteAttributes[lfn]

    attribute = self._retrieveRemoteAttribute(remotePath)
    self._remoteAttributes[lfn] = attribute
    return attribute

  def __getLocalAttribute(self, localPath):
    attribute = {}

    if not os.path.isfile(localPath):
      gLogger.debug('File not exists for getting attribute:', localPath)
      return attribute

    size = os.path.getsize(localPath)
    attribute['size'] = size

    mtime = os.path.getmtime(localPath)
    attribute['time'] = mtime

    return attribute


  def __localValid(self, lfn, localPath):
    remoteAttribute = self.__getRemoteAttribute(lfn)
    localAttribute = self.__getLocalAttribute(localPath)

    if not (remoteAttribute and localAttribute):
      return False

    if remoteAttribute['size'] != localAttribute['size']:
      return False

    if remoteAttribute['time'] != localAttribute['time']:
      return False

    return True

  def __setFileTime(self, localPath, mtime):
    os.utime(localPath, (mtime, mtime))

  def __removeLocal(self, localPath):
    if os.path.isfile(localPath):
      gLogger.debug('Remove invalid local file:', localPath)
      os.remove(localPath)

  def __parseMetadata(self, metadata):
    attribute = {}
    attribute['size'] = metadata.get('Size', 0)
    attribute['time'] = self.__utc2Local(metadata.get('ModificationDate', datetime.datetime(1900,1,1,0,0,0)))
    if self._useChecksum:
      attribute[lfn]['checksum'] = metadata.get('Checksum', '')
      attribute[lfn]['checksum_type'] = metadata.get('ChecksumType', '')
    return attribute

  def __utc2Local(self, utc_st):
    return time.mktime(utc_st.utctimetuple()) - time.timezone



from IHEPDIRAC.Badger.private.output.getfile.GetFile     import GetFile
from IHEPDIRAC.Badger.private.output.getfile.LocalMount  import LocalMount
from IHEPDIRAC.Badger.private.output.getfile.Rsync       import Rsync

class LocalRsyncGetFile(LocalMount, Rsync, GetFile):
  def __init__(self):
    super(LocalRsyncGetFile, self).__init__()



""" TestBase

  Base class for all tests.

"""

import threading
from abc   import ABCMeta,abstractmethod


__RCSID__ = '$Id:  $'

LOCK = threading.Lock()


class TestBase( object ):
  """
    TestBase is a simple base class for all tests. Real test classes should
    implement its doTest and getTestResult method.
  """

  __metaclass__ = ABCMeta

  def __init__( self, args = None, apis = None ):
    self.apis = apis or {}
    self.args = args or {}


  @abstractmethod
  def doTest( self, elementDict ):
    """
      to be extended by real tests.
    """

    return None



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# author: zhanggang

import threading
import time
import random

class LockedTierator(object):
  def __init__(self, it):
    self.lock = threading.Lock()
    self.it = it.__iter__()

  def __iter__(self): return self

  def next(self):
    self.lock.acquire()
    try:
      return self.it.next()
    finally:
      self.lock.release()

class IWorker(object):

  def get_file_list(self):
    raise NotImplementedError

  def Do(self, item):
    raise NotImplementedError

class MultiWorker(object):

  def __init__(self, worker, pool_size=5):
    self.max_pool_size = pool_size
    self.pool = []
    self.worker = worker
    self.worker.get_file_list = LockedTierator(self.worker.get_file_list())
    self.it = iter(self.worker.get_file_list)

  def check_pool(self):
    return self.max_pool_size - len(self.pool)

  def clear_pool(self):
    for t in self.pool[:]:
      if not t.isAlive():
        self.pool.remove(t)

  def get_file(self):
    try:
      return self.it.next()
    except StopIteration:
      return None
    pass

  def main(self):
    while True:
      self.clear_pool()
      residual = self.check_pool()
      if not residual:
        continue
      for i in range(residual):
        f = self.get_file()
        if f is None and len(self.pool) == 0:
          return

        t = threading.Thread( target=self.worker.Do, args=(f,) )
        self.pool.append(t)
        #t.setDaemon(True)
        t.start()

      #break


if __name__ == "__main__":

  worker = UploadWorker()
  mw = MultiWorker(worker)
  mw.main()



""" ResourceManagementIHEPClient

  ResourceManagementIHEPClient for IHEPDIRAC.

"""

from DIRAC                      import gLogger, S_ERROR
from datetime                   import datetime, timedelta
from DIRAC.Core.DISET.RPCClient import RPCClient

__RCSID__ = '$Id:  $'


def uppercase_first_letter(key):
  return key[0].upper() + key[1:]


class ResourceManagementIHEPClient( object ):

  def __init__( self , serviceIn = None ):
      '''
      The client tries to connect to :class:ResourceManagementIHEPDB by default. If it
      fails, then tries to connect to the Service :class:ResourceManagementIHEPHandler.
      '''

      if not serviceIn:
        self.gate = RPCClient( "ResourceStatus/ResourceManagementIHEP" )
      else:
        self.gate = serviceIn

  def _prepare(self, sendDict):

      # remove unnecessary key generated by locals()
      del sendDict['self']

      # make each key name uppercase to match database column names (case sensitive)
      for key, value in sendDict.items():
        del sendDict[key]

        # apply default values
        if key in ['dateEffective', 'lastCheckTime'] and value is None:
          sendDict.update({uppercase_first_letter(key): datetime.utcnow() })
        elif key == 'tokenExpiration' and value is None:
          sendDict.update({uppercase_first_letter(key): datetime.utcnow() + timedelta(hours=24)})
        else:
          sendDict.update({uppercase_first_letter(key): value})

      return sendDict


  # JobCache Methods ...........................................................

  def selectJobCache( self, site = None, maskStatus = None, efficiency = None,
                      running = None, waiting = None, done = None, failed = None, completed = None,
                      stalled = None, status = None, lastCheckTime = None, meta = None ):
    """
    Gets from JobCache all rows that match the parameters given.

    :Parameters:
      **site** - `[, string, list ]`
        name of the site element
      **maskStatus** - `[, string, list ]`
        maskStatus for the site
      **efficiency** - `[, float, list ]`
        job efficiency ( successful / total )
      **running** - `[, integer, list ]`
        number of running jobs
      **waiting** - `[, integer, list ]`
        number of waiting jobs
      **done** - `[, integer, list ]`
        number of done jobs
      **failed** - `[, integer, list ]`
        number of failed jobs
      **completed** - `[, integer, list ]`
        number of completed jobs
      **stalled** - `[, integer, list ]`
        number of stalled jobs
      **status** - `[, string, list ]`
        status for the site computed
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp setting last time the result was checked
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'select', 'JobCache', locals() )


  def deleteJobCache( self, site = None, maskStatus = None, efficiency = None,
                      running = None, waiting = None, done = None, failed = None, completed = None,
                      stalled = None, status = None, lastCheckTime = None, meta = None  ):
    """
    Deletes from JobCache all rows that match the parameters given.

    :Parameters:
      **site** - `[, string, list ]`
        name of the site element
      **maskStatus** - `[, string, list ]`
        maskStatus for the site
      **efficiency** - `[, float, list ]`
        job efficiency ( successful / total )
      **running** - `[, integer, list ]`
        number of running jobs
      **waiting** - `[, integer, list ]`
        number of waiting jobs
      **done** - `[, integer, list ]`
        number of done jobs
      **failed** - `[, integer, list ]`
        number of failed jobs
      **completed** - `[, integer, list ]`
        number of completed jobs
      **stalled** - `[, integer, list ]`
        number of stalled jobs
      **status** - `[, string, list ]`
        status for the site computed
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp setting last time the result was checked
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'delete', 'JobCache', locals() )


  def addOrModifyJobCache( self, site = None, maskStatus = None, efficiency = None,
                           running = None, waiting = None, done = None, failed = None, completed = None,
                           stalled = None, status = None, lastCheckTime = None, meta = None  ):
    """
    Adds or updates-if-duplicated to StorageCache. Using `site` to query
    the database, decides whether to insert or update the table.

    :Parameters:
      **site** - `[, string, list ]`
        name of the site element
      **maskStatus** - `[, string, list ]`
        maskStatus for the site
      **efficiency** - `[, float, list ]`
        job efficiency ( successful / total )
      **running** - `[, integer, list ]`
        number of running jobs
      **waiting** - `[, integer, list ]`
        number of waiting jobs
      **done** - `[, integer, list ]`
        number of done jobs
      **failed** - `[, integer, list ]`
        number of failed jobs
      **completed** - `[, integer, list ]`
        number of completed jobs
      **stalled** - `[, integer, list ]`
        number of stalled jobs
      **status** - `[, string, list ]`
        status for the site computed
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp setting last time the result was checked
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    meta = { 'onlyUniqueKeys' : True }
    return self._query( 'addOrModify', 'JobCache', locals() )


  # StorageCache methods ......................................................

  def selectStorageCache( self, sE = None, occupied = None, free = None,
                          usage = None, lastCheckTime = None, meta = None ):
    """
    Gets from StorageCache all rows that match the parameters given.

    :Parameters:
      **sE** - `[, string, list ]`
        name of se
      **occupied** - `[, integer, list ]`
        occupied storage of se
      **free** - `[, integer, list ]`
        free storage of se
      **usage** - `[, float, list ]`
        usage rate of se
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'select', 'StorageCache', locals() )

  def deleteStorageCache( self, sE = None, occupied = None, free = None,
                          usage = None, lastCheckTime = None, meta = None ):
    """
    Deletes from StorageCache all rows that match the parameters given.

     :Parameters:
      **sE** - `[, string, list ]`
        name of se
      **occupied** - `[, integer, list ]`
        occupied storage of se
      **free** - `[, integer, list ]`
        free storage of se
      **usage** - `[, float, list ]`
        usage rate of se
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'delete', 'StorageCache', locals() )

  def addOrModifyStorageCache( self, sE = None, occupied = None, free = None,
                               usage = None, lastCheckTime = None, meta =None ):
    """
    Adds or updates-if-duplicated to StorageCache. Using `sE` to query
    the database, decides whether to insert or update the table.

     :Parameters:
      **sE** - `[, string, list ]`
        name of se
      **occupied** - `[, integer, list ]`
        occupied storage of se
      **free** - `[, integer, list ]`
        free storage of se
      **usage** - `[, float, list ]`
        usage rate of se
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    meta = { 'onlyUniqueKeys' : True }
    return self._query( 'addOrModify', 'StorageCache', locals() )


  # WorkNodeCache methods ......................................................

  def selectWorkNodeCache( self, host = None, site = None, done = None,
                       failed = None, efficiency = None, lastCheckTime = None, meta = None ):
    """
    Gets from WorkNodeCache all rows that match the parameters given.

    :Parameters:
      **host** - `[, string, list ]`
        name of the host
      **site** - `[, string, list ]`
        name of the site
      **done** - `[, integer, list ]`
        number for the done jobs
      **failed** - `[, integer, list ]`
        number for the failed jobs
      **efficiency** - `[, float, list ]`
        job efficiency
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'select', 'WorkNodeCache', locals() )


  def deleteWorkNodeCache( self, host = None, site = None, done = None,
                       failed = None, efficiency = None, lastCheckTime = None, meta = None ):
    """
    Deletes from WorkNodeCache all rows that match the parameters given.

    :Parameters:
      **host** - `[, string, list ]`
        name of the host
      **site** - `[, string, list ]`
        name of the site
      **done** - `[, integer, list ]`
        number for the done jobs
      **failed** - `[, integer, list ]`
        number for the failed jobs
      **efficiency** - `[, float, list ]`
        job efficiency
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'delete', 'WorkNodeCache', locals() )


  def addOrModifyWorkNodeCache( self, host = None, site = None,  done = None,
                       failed = None, efficiency = None, lastCheckTime = None, meta = None ):
    """
    Adds or updates-if-duplicated to WorkNodeCache. Using `host` to query
    the database, decides whether to insert or update the table.

    :Parameters:
      **host** - `[, string, list ]`
        name of the host
      **site** - `[, string, list ]`
        name of the site
      **done** - `[, integer, list ]`
        number for the done jobs
      **failed** - `[, integer, list ]`
        number for the failed jobs
      **efficiency** - `[, float, list ]`
        job efficiency
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    meta = { 'onlyUniqueKeys' : True }
    return self._query( 'addOrModify', 'WorkNodeCache', locals() )


  # SAMResult methods ......................................................

  def selectSAMResult( self, elementName = None, testType = None, elementType = None,
                          status = None, log = None, jobID = None, submissionTime = None,
                          completionTime = None, applicationTime = None, lastCheckTime = None, meta = None ):
    """
    Gets from SAMResult all rows that match the parameters given.

    :Parameters:
      **elementName** - `[, string, list ]`
        name of the element
      **testType** - `[, string, list ]`
        type of the test
      **elementType** - `[, string, list ]`
        type of the element
      **status** - `[, string, list ]`
        status for the executed test
      **log** - `[, string, list ]`
        log for the executed test
      **jobID** - `[, integer, list ]`
        the test job id
      **submissionTime** - `[, datetime, list ]`
        submission time for the test job
      **completionTime** - `[, datetime, list ]`
        completion time for the test job
      **applicationTime** - `[, float, list ]`
        real running time for the test
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'select', 'SAMResult', locals() )


  def deleteSAMResult( self, elementName = None, testType = None, elementType = None,
                          status = None, log = None, jobID = None, submissionTime = None,
                          completionTime = None, applicationTime = None, lastCheckTime = None, meta = None ):
    """
    Deletes from SAMResult all rows that match the parameters given.

    :Parameters:
      **elementName** - `[, string, list ]`
        name of the element
      **testType** - `[, string, list ]`
        type of the test
      **elementType** - `[, string, list ]`
        type of the element
      **status** - `[, string, list ]`
        status for the executed test
      **log** - `[, string, list ]`
        log for the executed test
      **jobID** - `[, integer, list ]`
        the test job id
      **submissionTime** - `[, datetime, list ]`
        submission time for the test job
      **completionTime** - `[, datetime, list ]`
        completion time for the test job
      **applicationTime** - `[, float, list ]`
        real running time for the test
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'delete', 'SAMResult', locals() )


  def addOrModifySAMResult( self, elementName = None, testType = None, elementType = None,
                               status = None, log = None, jobID = None, submissionTime = None,
                               completionTime = None, applicationTime = None, lastCheckTime = None, meta = None ):
    """
    Adds or updates-if-duplicated to SAMResult. Using `elementName`
    and `testType` to query the database, decides whether to insert or update
    the table.

    :Parameters:
      **elementName** - `[, string, list ]`
        name of the element
      **testType** - `[, string, list ]`
        type of the test
      **elementType** - `[, string, list ]`
        type of the element
      **status** - `[, string, list ]`
        status for the executed test
      **log** - `[, string, list ]`
        log for the executed test
      **jobID** - `[, integer, list ]`
        the test job id
      **submissionTime** - `[, datetime, list ]`
        submission time for the test job
      **completionTime** - `[, datetime, list ]`
        completion time for the test job
      **applicationTime** - `[, float, list ]`
        real running time for the test
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    meta = { 'onlyUniqueKeys' : True }
    return self._query( 'addOrModify', 'SAMResult', locals() )


  # ResourceSAMStatus methods ......................................................

  def selectResourceSAMStatus( self, vO = None, elementName = None,
                               elementType = None, tests = None, status = None,
                               lastCheckTime = None, meta = None ):
    """
    Gets from ResourceSAMStatus all rows that match the parameters given.

    :Parameters:
      **vO** - `[, string, list]`
        name of the vo
      **elementName** - `[, string, list ]`
        name of the element
      **elementType** - `[, string, list ]`
        type of the element
      **tests** - `[, string, list ]`
        tests which executed for the element
      **status** - `[, string, list ]`
        sam status of the element
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'select', 'ResourceSAMStatus', locals() )


  def deleteResourceSAMStatus( self, vO = None, elementName = None,
                               elementType = None, tests = None, status = None,
                               lastCheckTime = None, meta = None ):
    """
    Deletes from ResourceSAMStatus all rows that match the parameters given.

    :Parameters:
      **vO** - `[, string, list]`
        name of the vo
      **elementName** - `[, string, list ]`
        name of the element
      **elementType** - `[, string, list ]`
        type of the element
      **tests** - `[, string, list ]`
        tests which executed for the element
      **status** - `[, string, list ]`
        sam status of the element
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'delete', 'ResourceSAMStatus', locals() )


  def addOrModifyResourceSAMStatus( self, vO = None, elementName = None,
                                    elementType = None, tests = None, status = None,
                                    lastCheckTime = None,meta = None ):
    """
    Adds or updates-if-duplicated to ResourceSAMStatus. Using `elementName`
    to query the database, decides whether to insert or update the table.

    :Parameters:
      **vO** - `[, string, list]`
        name of the vo
      **elementName** - `[, string, list ]`
        name of the element
      **elementType** - `[, string, list ]`
        type of the element
      **tests** - `[, string, list ]`
        tests which executed for the element
      **status** - `[, string, list ]`
        sam status of the element
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    meta = { 'onlyUniqueKeys' : True }
    return self._query( 'addOrModify', 'ResourceSAMStatus', locals() )


  # SiteSAMStatus methods ......................................................

  def selectSiteSAMStatus( self, vO = None, site = None, siteType = None, status = None,
                                cEStatus = None, sEStatus = None, lastCheckTime = None, meta = None ):
    """
    Gets from SiteSAMStatus all rows that match the parameters given.

    :Parameters:
      **vO** - `[, string, list]`
        name of the vo
      **site** - `[, string, list ]`
        name of the site
      **siteType** - `[, string, list ]`
        type of the site
      **status** - `[, string, list ]`
        sam status of the site
      **cEStatus** - `[, string, list ]`
        ce sam status of the site
      **sEStatus** - `[, string, list ]`
        se sam status of the site
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'select', 'SiteSAMStatus', locals() )


  def deleteSiteSAMStatus( self, vO = None, site = None, siteType = None, status = None,
                                cEStatus = None, sEStatus = None, lastCheckTime = None, meta = None ):
    """
    Deletes from SiteSAMStatus all rows that match the parameters given.

    :Parameters:
      **vO** - `[, string, list]`
        name of the vo
      **site** - `[, string, list ]`
        name of the site
      **siteType** - `[, string, list ]`
        type of the site
      **status** - `[, string, list ]`
        sam status of the site
      **cEStatus** - `[, string, list ]`
        ce sam status of the site
      **sEStatus** - `[, string, list ]`
        se sam status of the site
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    return self._query( 'delete', 'SiteSAMStatus', locals() )


  def addOrModifySiteSAMStatus( self, vO = None, site = None, siteType = None, status = None,
                                     cEStatus = None, sEStatus = None, lastCheckTime = None, meta = None ):
    """
    Adds or updates-if-duplicated to SiteSAMStatus. Using `site`
    to query the database, decides whether to insert or update the table.

    :Parameters:
      **vO** - `[, string, list]`
        name of the vo
      **site** - `[, string, list ]`
        name of the site
      **siteType** - `[, string, list ]`
        type of the site
      **status** - `[, string, list ]`
        sam status of the site
      **cEStatus** - `[, string, list ]`
        ce sam status of the site
      **sEStatus** - `[, string, list ]`
        se sam status of the site
      **lastCheckTime** - `[, datetime, list ]`
        time-stamp from which the result is effective
      **meta** - `[, dict ]`
        meta-data for the MySQL query. It will be filled automatically with the\
       `table` key and the proper table name.
    """

    meta = { 'onlyUniqueKeys' : True }
    return self._query( 'addOrModify', 'SiteSAMStatus', locals() )


  # get history methods ......................................................

  def selectSAMResultLog( self, elementName = None, testType = None, elementType = None,
                          status = None, log = None, jobID = None, submissionTime = None,
                          completionTime = None, applicationTime = None, lastCheckTime = None, meta = None ):
    return self._query( 'select', 'SAMResultLog', locals() )


  def selectResourceSAMStatusLog( self, vO = None, elementName = None, elementType = None,
                                         tests = None, status = None, lastCheckTime = None, meta = None ):
    return self._query( 'select', 'ResourceSAMStatusLog', locals() )


  def selectSiteSAMStatusLog( self, vO = None, site = None, siteType = None, status = None,
                                cEStatus = None, sEStatus = None, lastCheckTime = None, meta = None ):
    return self._query( 'select', 'SiteSAMStatusLog', locals() )


  # Protected methods ..........................................................

  def _query( self, queryType, tableName, parameters ):
    '''
    It is a simple helper, this way inheriting classes can use it.
    '''

    return self.__query( queryType, tableName, parameters )


  def __query( self, queryType, tableName, parameters ):
    '''
      This method is a rather important one. It will format the input for the DB
      queries, instead of doing it on a decorator. Two dictionaries must be passed
      to the DB. First one contains 'columnName' : value pairs, being the key
      lower camel case. The second one must have, at lease, a key named 'table'
      with the right table name.
    '''
    # Functions we can call, just a light safety measure.
    _gateFunctions = [ 'insert', 'update', 'select', 'delete', 'addOrModify', 'addIfNotThere'  ]
    if not queryType in _gateFunctions:
      return S_ERROR( '"%s" is not a proper gate call' % queryType )

    gateFunction = getattr( self.gate, queryType )

    # If meta is None, we set it to {}
    meta = ( True and parameters.pop( 'meta' ) ) or {}
#    params = parameters
    # Remove self, added by locals()
    del parameters[ 'self' ]

    meta[ 'table' ] = tableName

    gLogger.debug( 'Calling %s, with \n params %s \n meta %s' % ( queryType, parameters, meta ) )
    return gateFunction( parameters, meta )


#...............................................................................
#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF
